from typing import Any

try:
    from openeye.oechem import OECalculateMolecularWeight
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import to_digit, to_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The molecular weight of a given molecule.

By default, all atoms are assumed their average atomic weight.
"""


class MolecularWeight(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        digitise: bool = False,
        start: int = 0,
        stop: int = 700,
        num: int = 14,
        isotopic: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[float]]:
        """Calculates the molecular weight of each input molecule.

        By default, all atoms are assumed their average atomic weight.

        Args:
            samples: The sample molecules to featurise.
            digitise: Whether to digitise the results. Defaults to `False`.
            start: If `digitise=True`, the start of the binning range. Defaults
                to 0.
            stop: If `digitise=True`, the end of the binning range. Defaults
                to 700.
            num: If `digitise=True`, the number of bins. Defaults to 14.
            isotopic: Whether to use the `OEGetIsotopicWeight` function instead
                of `OEGetAverageWeight`. If `True`, atoms must have a specified
                (non-zero) isotopic mass, as returned by
                `OEAtomBase.GetIsotope`. Defaults to `False`.

        Returns:
            The molecular weight of each input molecule.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('molecular_weight')
            >>> samples = ['c1ccccc1', 'C']
            >>> representation.featurise(samples)
            {'molecular_weight': [78.11184, 16.04246]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        results = []
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                result = OECalculateMolecularWeight(mol, isotopic=isotopic)

                if digitise:
                    result = to_digit(result, start=start, stop=stop, num=num).tolist()

                results.append(result)

        return {self.tag: results}



import pytest


def test_invalid_kwarg_raises_value_error(fixture_mock_splitting_strategy):
    """That attempting split with invalid keyword arguments raises."""
    strategy = fixture_mock_splitting_strategy
    data = ["CCCC", "CC"] * 10
    with pytest.raises(ValueError, match=r"Unknown split parameter\(s\)"):
        next(strategy.split(data, invalid_kwarg=True))



"""Top-k accuracy classification score."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import top_k_accuracy_score

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
This metric computes the number of times where the correct label is among the
top k labels predicted (ranked by predicted scores). Note that the multilabel
case isn`t covered here.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Target scores. These can be either probability estimates or
        non-thresholded decision values. The binary case expects scores with
        shape (n_samples,) while the multiclass case expects scores with shape
        (n_samples, n_classes). In the multiclass case, the order of the class
        scores must correspond to the order of ``labels``, if provided, or else
        to the numerical or lexicographical order of the labels in
        ``references``. If ``references`` does not contain all the labels,
        ``labels`` must be provided.
    references: Ground truth (correct) target values.
    k (optional): Number of most likely outcomes considered to find the correct
        label. Defaults to 2.
    normalize (optional): If False, return the number of correctly classified
        samples. Otherwise, return the fraction of correctly classified samples.
    sample_weight (optional): Sample weights.
    labels (optional): Multiclass only. List of labels that index the classes
        in ``predictions``. If ``None``, the numerical or lexicographical order
        of  the labels in ``references`` is used. If ``predictions`` does not
        contain all the labels, ``labels`` must be provided.

Returns:
    top_k_accuracy: The top-k accuracy score. The best performance is 1 with
    `normalize == True` and the number of samples with `normalize == False`.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("top_k_accuracy", config_name="multiclass")
    >>> references = [0, 1, 2, 2]
    >>> predictions = [[0.5, 0.2, 0.2],  # 0 is in top 2
    ...                [0.3, 0.4, 0.2],  # 1 is in top 2
    ...                [0.2, 0.4, 0.3],  # 2 is in top 2
    ...                [0.7, 0.2, 0.1]] # 2 isn't in top 2
    >>> metric.compute(references=references, predictions=predictions, k=2)
    {'top_k_accuracy': 0.75}
    >>> # Not normalizing gives the number of "correctly" classified samples
    >>> metric.compute(references=references, predictions=predictions, k=2, normalize=False)
    {'top_k_accuracy': 3}
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class TopKAccuracy(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("float")),
                    "references": datasets.Value("int32"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.top_k_accuracy_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        labels: list[str] | None = None,
        k: int = 2,
        normalize: bool = True,
        **kwargs: Any,
    ) -> MetricResult:
        score = top_k_accuracy_score(
            y_true=references,
            y_score=predictions,
            k=k,
            normalize=normalize,
            sample_weight=sample_weight,
            labels=labels,
        )
        return {self.tag: score}



import numpy as np
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.regression.proportion_within_fold import ProportionWithinFold


@pytest.fixture()
def fixture_metric():
    return load_metric("proportion_within_fold")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "proportion_within_fold" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, ProportionWithinFold)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_default_compute(fixture_metric):
    """That default scoring."""
    metric = fixture_metric
    predictions = [2.5, 4.0, 2, 8]
    references = [3, 0.5, 2, 7]
    result = metric.compute(predictions=predictions, references=references)
    assert pytest.approx(result["proportion_within_fold"], 0.1) == 0.75


def test_raises_warning_for_negative_inputs(fixture_metric):
    """Negative inputs should raise a warning as will need to be log transformed."""
    metric = fixture_metric
    predictions = [2.5, -4.0, 2, 8]
    references = [3, 0.5, 2, 7]
    with pytest.raises(RuntimeError):
        metric.compute(predictions=predictions, references=references)


def test_log_scale_gives_result(fixture_metric):
    """Transforming by log10 should give the same result as using the log_scale option."""
    metric = fixture_metric
    predictions = [2.5, 4.0, 2, 8]
    references = [3, 0.5, 2, 7]
    result = metric.compute(predictions=predictions, references=references)
    result_log_scaled = metric.compute(
        predictions=np.log10(predictions),
        references=np.log10(references),
        log_scaled_inputs=True,
    )
    assert (
        pytest.approx(result["proportion_within_fold"], 0.0001)
        == result_log_scaled["proportion_within_fold"]
    )



"""
Tests for API features advertised as part of the Representations protocol.
"""

from molflux.features.representation import Representations


def test_has_add_representation():
    """That Representations has add_representation method."""
    assert hasattr(Representations, "add_representation")
    assert callable(Representations.add_representation)


def test_has_featurise():
    """That Representations has a featurise method."""
    assert hasattr(Representations, "featurise")
    assert callable(Representations.featurise)



import tempfile
from pathlib import Path
from typing import Any

import molflux.modelzoo as mz


def load_from_dvc(repo_url: str, rev: str, model_path_in_repo: str) -> Any:
    try:
        from dvc.api import DVCFileSystem  # pyright: ignore
    except ImportError as err:
        from molflux.modelzoo.errors import OptionalDependencyImportError

        raise OptionalDependencyImportError("DVC", "dvc[s3]") from err
    else:
        fs = DVCFileSystem(repo_url, rev=rev, subrepos=True)
        with tempfile.TemporaryDirectory() as tmpdir:
            fs.download(
                model_path_in_repo,
                str(Path(tmpdir) / "model"),
                recursive=True,
            )
            model = mz.load_from_store(Path(tmpdir) / "model")

        return model


def load_from_dvc_or_disk(
    path: str | None = None,
    repo_url: str | None = None,
    rev: str | None = None,
    model_path_in_repo: str | None = None,
) -> Any:
    if path is not None:
        model = mz.load_from_store(path)
    elif (
        (repo_url is not None)
        and (rev is not None)
        and (model_path_in_repo is not None)
    ):
        model = load_from_dvc(
            repo_url=repo_url,
            rev=rev,
            model_path_in_repo=model_path_in_repo,
        )
    else:
        raise ValueError(
            "Must specify either 'path' or all of 'repo_url', 'rev', 'model_path_in_repo' (but not both).",
        )

    return model



from dataclasses import asdict
from typing import Any, Literal

import numpy as np
from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.pyod import (
    PyODClassificationMixin,
    PyODModelBase,
    PyODModelConfig,
)

try:
    from pyod.models.pca import PCA
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None


_DESCRIPTION = """
Principal component analysis (PCA) can be used in detecting outliers.
PCA is a linear dimensionality reduction using Singular Value Decomposition
of the data to project it to a lower dimensional space.

In this procedure, covariance matrix of the data can be decomposed to
orthogonal vectors, called eigenvectors, associated with eigenvalues. The
eigenvectors with high eigenvalues capture most of the variance in the
data.

Therefore, a low dimensional hyperplane constructed by k eigenvectors can
capture most of the variance in the data. However, outliers are different
from normal data points, which is more obvious on the hyperplane
constructed by the eigenvectors with small eigenvalues.

Therefore, outlier scores can be obtained as the sum of the projected
distance of a sample on all eigenvectors.
See :cite:`shyu2003novel,aggarwal2015outlier` for details.

Score(X) = Sum of weighted euclidean distance between each sample to the
hyperplane constructed by the selected eigenvectors
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_components : int, float, None or string
    Number of components to keep.
    if n_components is not set all components are kept::

        n_components == min(n_samples, n_features)

    if n_components == 'mle' and svd_solver == 'full', Minka\'s MLE is used
    to guess the dimension
    if ``0 < n_components < 1`` and svd_solver == 'full', select the number
    of components such that the amount of variance that needs to be
    explained is greater than the percentage specified by n_components
    n_components cannot be equal to n_features for svd_solver == 'arpack'.

n_selected_components : int, optional (default=None)
    Number of selected principal components
    for calculating the outlier scores. It is not necessarily equal to
    the total number of the principal components. If not set, use
    all principal components.

contamination : float in (0., 0.5), optional (default=0.1)
    The amount of contamination of the data set, i.e.
    the proportion of outliers in the data set. Used when fitting to
    define the threshold on the decision function.

inplace : bool (default False)
    If True, data passed to fit are overwritten and running
    fit(X).transform(X) will not yield the expected results,
    use fit_transform(X) instead.

whiten : bool, optional (default False)
    When True (False by default) the `components_` vectors are multiplied
    by the square root of n_samples and then divided by the singular values
    to ensure uncorrelated outputs with unit component-wise variances.

    Whitening will remove some information from the transformed signal
    (the relative variance scales of the components) but can sometime
    improve the predictive accuracy of the downstream estimators by
    making their data respect some hard-wired assumptions.

svd_solver : string {'auto', 'full', 'arpack', 'randomized'}
    auto :
        the solver is selected by a default policy based on `X.shape` and
        `n_components`: if the input data is larger than 500x500 and the
        number of components to extract is lower than 80% of the smallest
        dimension of the data, then the more efficient 'randomized'
        method is enabled. Otherwise the exact full SVD is computed and
        optionally truncated afterwards.
    full :
        run exact full SVD calling the standard LAPACK solver via
        `scipy.linalg.svd` and select the components by postprocessing
    arpack :
        run SVD truncated to n_components calling ARPACK solver via
        `scipy.sparse.linalg.svds`. It requires strictly
        0 < n_components < X.shape[1]
    randomized :
        run randomized SVD by the method of Halko et al.

tol : float >= 0, optional (default .0)
    Tolerance for singular values computed by svd_solver == 'arpack'.

iterated_power : int >= 0, or 'auto', (default 'auto')
    Number of iterations for the power method computed by
    svd_solver == 'randomized'.

random_state : int, np.random.Generator instance or None, optional (default None)
    If int, random_state is the seed used by the random number generator;
    If Generator instance, random_state is the random number generator;
    If None, the random number generator is the Generator instance used
    by `np.random`. Used when ``svd_solver`` == 'arpack' or 'randomized'.

weighted : bool, optional (default=True)
    If True, the eigenvalues are used in score computation.
    The eigenvectors with small eigenvalues comes with more importance
    in outlier score calculation.

standardisation : bool, optional (default=True)
    If True, perform standardization first to convert
    data to zero mean and unit variance.
    See http://scikit-learn.org/stable/auto_examples/preprocessing/plot_scaling_importance.html
"""

Auto = Literal["auto"]
SVDSolver = Literal["auto", "full", "arpack", "randomized"]


class Config:
    extra = "forbid"
    arbitrary_types_allowed = True
    smart_union = True


@dataclasses.dataclass(config=Config)
class PCADetectorConfig(PyODModelConfig):
    n_components: int | float | str | None = None
    n_selected_components: int | None = None
    contamination: float = 0.1
    inplace: bool = False
    whiten: bool = False
    svd_solver: SVDSolver = "auto"
    tol: float = 0
    iterated_power: int | Auto = "auto"
    random_state: int | None | np.random.Generator = None
    weighted: bool = True
    standardisation: bool = True


class PCADetector(PyODClassificationMixin, PyODModelBase[PCADetectorConfig]):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[PCADetectorConfig]:
        return PCADetectorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> PCA:
        config = self.model_config
        return PCA(
            n_components=config.n_components,
            n_selected_components=config.n_selected_components,
            contamination=config.contamination,
            copy=not config.inplace,
            whiten=config.whiten,
            svd_solver=config.svd_solver,
            tol=config.tol,
            iterated_power=config.iterated_power,
            random_state=config.random_state,
            weighted=config.weighted,
            standardization=config.standardisation,
        )



import json

import pytest
from openeye import oechem

from molflux.features import (
    Representation,
    list_representations,
    load_from_dict,
    load_representation,
)
from molflux.features.errors import FeaturisationError
from molflux.features.representations.openeye.sd.sd_feature import (
    DType,
    SDFeature,
)

representation_name = "sd_feature"


@pytest.fixture(scope="module")
def fixture_molecules() -> list[oechem.OEMolBase]:
    """Fixture to obtain three molecules."""
    smiles = ["c1ccccc1", "c1(CCCC)ccccc1", "CC"]
    mols = [oechem.OEGraphMol() for _ in range(len(smiles))]
    for mol, smi in zip(mols, smiles, strict=False):
        oechem.OESmilesToMol(mol, smi)
    return mols


@pytest.fixture(scope="module")
def fixture_on_bits() -> list[list[int]]:
    return [[3, 6, 8, 2], [9, 5, 3, 1, 8, 7], [0, 4]]


@pytest.fixture(scope="module")
def fixture_dense_bits() -> list[list[int]]:
    return [
        [0, 0, 1, 1, 0, 0, 1, 0, 1, 0],
        [0, 1, 0, 1, 0, 1, 0, 1, 1, 1],
        [1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
    ]


@pytest.fixture(scope="module")
def fixture_molecules_sd_features(
    fixture_molecules: list[oechem.OEMolBase],
    fixture_on_bits: list[list[int]],
    fixture_dense_bits: list[list[int]],
) -> list[oechem.OEMolBase]:
    """Fixture to append SD tags for particular features onto the sampled molecules."""
    molecules_with_sd_tags = []
    for org_mol, on_bits, dense_bits in zip(
        fixture_molecules,
        fixture_on_bits,
        fixture_dense_bits,
        strict=False,
    ):
        mol = org_mol.CreateCopy()
        # simulate newline separated 'on' bits.
        oechem.OESetSDData(
            mol,
            "newline-on-bits",
            "\n".join(
                [*map(str, on_bits)],
            ),
        )
        # simulate comma separated 'on' bits.
        oechem.OESetSDData(
            mol,
            "comma-on-bits",
            ",".join([*map(str, on_bits)]),
        )
        # simulate JSON dense list.
        oechem.OESetSDData(
            mol,
            "json-dense-array",
            json.dumps(dense_bits),
        )
        # simulate JSON sparse list.
        oechem.OESetSDData(
            mol,
            "json-sparse-bits",
            json.dumps(on_bits),
        )
        molecules_with_sd_tags.append(mol)
    return molecules_with_sd_tags


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class():
    """That the catalogue name is mapped to the appropriate class."""
    representation = load_representation(representation_name)
    assert isinstance(representation, SDFeature)


def test_implements_protocol():
    """That the representation implements the public Representation protocol."""
    representation = load_representation(representation_name)
    assert isinstance(representation, Representation)


@pytest.mark.parametrize(
    "sd_representation",
    [
        {
            "name": "sd_feature",
            "presets": {"sd_tags": ["newline-on-bits"], "as_json": False},
        },
        {"name": "sd_feature", "presets": {"sd_tags": ["json-dense-array"]}},
        {"name": "sd_feature", "presets": {"sd_tags": ["json-sparse-bits"]}},
        {
            "name": "sd_feature",
            "presets": {"sd_tags": ["json-sparse-bits", "json-dense-array"]},
        },
        {
            "name": "sd_feature",
            "presets": {
                "sd_tags": [
                    "json-sparse-bits",
                    "json-dense-array",
                    "newline-on-bits",
                ],
                "as_json": ["json-sparse-bits", "json-dense-array"],
            },
        },
        {
            "name": "sd_feature",
            "presets": {
                "sd_tags": [
                    "json-sparse-bits",
                    "json-dense-array",
                    "newline-on-bits",
                ],
                "as_json": ["json-sparse-bits", "json-dense-array"],
                "dtype": {"newline-on-bits": "float"},
            },
        },
        {
            "name": "sd_feature",
            "presets": {
                "sd_tags": [
                    "comma-on-bits",
                    "newline-on-bits",
                ],
                "as_json": False,
                "dtype": {"newline-on-bits": "float"},
                "separator": {"comma-on-bits": ","},
            },
        },
    ],
    ids=[
        "newline-separated-features",
        "json-dense-array",
        "json-sparse-array",
        "multiple-json",
        "multiple-mix",
        "dtype-cast",
        "different-separators",
    ],
)
def test_featurises_with_different_sd_tag_parsing(
    sd_representation,
    fixture_molecules_sd_features,
):
    """That the featurisation of an SDFeature representation
    works as expected with different sd tags and parsers used."""
    representation = load_from_dict(sd_representation)

    assert "sd_tags" in representation.state
    assert representation.state["sd_tags"]
    # Obtain features.
    features = representation.featurise(fixture_molecules_sd_features)

    # Check all features are present.
    assert {
        f"{representation.tag}::{sd_tag}" for sd_tag in representation.state["sd_tags"]
    } == {*features}

    # Checks for individual feature types.
    if f"{representation.tag}::newline-on-bits" in features:
        assert isinstance(features[f"{representation.tag}::newline-on-bits"][0], list)
        rep_dtype = representation.state["dtype"]
        dtype = (
            rep_dtype.get("newline-on-bits", "str")
            if isinstance(rep_dtype, dict)
            else rep_dtype
        )
        dtype_type = DType[dtype.upper()].value

        assert isinstance(
            features[f"{representation.tag}::newline-on-bits"][0][0],
            dtype_type,
        )
    if f"{representation.tag}::comma-on-bits" in features:
        assert isinstance(features[f"{representation.tag}::comma-on-bits"][0], list)
        rep_dtype = representation.state["dtype"]
        dtype = (
            rep_dtype.get("comma-on-bits", "str")
            if isinstance(rep_dtype, dict)
            else rep_dtype
        )
        dtype_type = DType[dtype.upper()].value
        assert isinstance(
            features[f"{representation.tag}::comma-on-bits"][0][0],
            dtype_type,
        )
    if f"{representation.tag}::json-dense-array" in features:
        assert isinstance(features[f"{representation.tag}::json-dense-array"][0], list)
        assert isinstance(
            features[f"{representation.tag}::json-dense-array"][0][0],
            int,
        )

    if f"{representation.tag}::json-sparse-bits" in features:
        assert isinstance(features[f"{representation.tag}::json-sparse-bits"][0], list)
        assert isinstance(
            features[f"{representation.tag}::json-sparse-bits"][0][0],
            int,
        )


def test_raises_if_sd_tags_not_present(fixture_molecules):
    """That SD feature featurisation fails without SD tags being present."""
    sd_representation = {
        "name": "sd_feature",
        "presets": {"sd_tags": ["json-dense-array"]},
    }
    representation = load_from_dict(sd_representation)
    with pytest.raises(FeaturisationError):
        representation.featurise(fixture_molecules)

    # double check the cause of the error is what it should be
    # (this is an internal consistency check only, not an invariant of the code)
    try:
        representation.featurise(fixture_molecules)
    except FeaturisationError as e:
        assert isinstance(e.__cause__, ValueError)
        assert "Unable to find SD tag" in str(e.__cause__)


def test_failure_if_no_sd_tags_supplied(fixture_molecules_sd_features):
    """Ensures an error is thrown if no SD tags are specified."""
    representation = load_representation(representation_name)
    # No SD tags supplied.
    assert not representation.state.get("sd_tags", None)
    with pytest.raises(ValueError, match="Must supply SD tags .*"):
        representation.featurise(fixture_molecules_sd_features)



from typing import Any

try:
    from rdkit.Chem import RDKFingerprint

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
The RDKit topological fingerprint for a molecule.

These are topological (Daylight like) fingerprints for a molecule using an
alternate (faster) hashing algorithm. This RDKit-specific fingerprint is
inspired by (though it differs significantly from) public descriptions of the
Daylight fingerprint 7. The fingerprinting algorithm identifies all subgraphs
in the molecule within a particular range of sizes, hashes each subgraph to
generate a raw bit ID, mods that raw bit ID to fit in the assigned fingerprint
size, and then sets the corresponding bit. Options are available to generate
count-based forms of the fingerprint or “non-folded” forms (using a sparse
representation).

The default scheme for hashing subgraphs is to hash the individual bonds based
on:
* the types of the two atoms. Atom types include the atomic number (mod 128), and whether or not the atom is aromatic.
* the degrees of the two atoms in the path.
* the bond type (or AROMATIC if the bond is marked as aromatic)
"""


class Topological(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        min_path: int = 1,
        max_path: int = 7,
        fp_size: int = 2048,
        n_bits_per_hash: int = 2,
        use_hs: bool = True,
        tgt_density: float = 0,
        min_size: int = 128,
        branched_paths: bool = True,
        use_bond_order: bool = True,
        atom_invariants: list[int] | None = None,
        from_atoms: list | None = None,
        atom_bits: list | None = None,
        bit_info: dict | None = None,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates topological (Daylight like) fingerprints for each input
        molecules.

        Args:
            samples: The molecules to be fingerprinted.
            min_path: The minimum path length (in bonds) to be included.
                Defaults to `1`.
            max_path: The maximum path length (in bonds) to be included.
                Defaults to `7`
            fp_size: The size of the fingerprint. Defaults to `2048`.
            n_bits_per_hash: The number of bits to be set by each path.
                Defaults to `2`.
            use_hs: Toggles inclusion of Hs in paths (if the molecule has
                explicit Hs). Defaults to `True`.
            tgt_density: If the generated fingerprint is below this density,
                it will be folded until the density is reached. Defaults to `0`.
            min_size: The minimum size to which the fingerprint will be folded.
                Defaults to `128`.
            branched_paths: Toggles generation of branched subgraphs, not just
                linear paths. Defaults to `True`.
            use_bond_order: Toggles inclusion of bond orders in the path hashes.
                Defaults to `True`.
            atom_invariants: A vector of atom invariants to use while hashing
                the paths. Defaults to `None`.
            from_atoms: Only paths starting at these atoms will be included.
                Defaults to `None`.
            atom_bits: Used to return the bits that each atom is involved in
                (should be at least mol.numAtoms long). Defaults to `None`.
            bit_info: Defaults to `None`.

        Returns:
            Topological (Daylight like) fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('topological')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples, fp_size=16)
            {'topological': [[0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        topo_fp_list: list[list] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)
                rd_fp = RDKFingerprint(
                    mol,
                    minPath=min_path,
                    maxPath=max_path,
                    fpSize=fp_size,
                    nBitsPerHash=n_bits_per_hash,
                    useHs=use_hs,
                    tgtDensity=tgt_density,
                    minSize=min_size,
                    branchedPaths=branched_paths,
                    useBondOrder=use_bond_order,
                    atomInvariants=atom_invariants or [],
                    fromAtoms=from_atoms or [],
                    atomBits=atom_bits or [],
                    bitInfo=bit_info or {},
                )

                if rd_fp.GetNumOnBits() == 0:
                    topo_fp_list.append([0] * rd_fp.GetNumBits())
                else:
                    topo_fp_list.append(rd_fp.ToList())

        return {self.tag: topo_fp_list}



from collections.abc import Callable
from dataclasses import asdict
from typing import Any, Literal

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.models.pyod import (
    PyODClassificationMixin,
    PyODModelBase,
    PyODModelConfig,
)

try:
    from pyod.models.ocsvm import OCSVM
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("pyod", e) from None


_DESCRIPTION = """
Wrapper of scikit-learn one-class SVM Class with more functionalities.
Unsupervised Outlier Detection.

Estimate the support of a high-dimensional distribution.

The implementation is based on libsvm.
See http://scikit-learn.org/stable/modules/svm.html#svm-outlier-detection
and :cite:`scholkopf2001estimating`.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
kernel : string, optional (default='rbf')
     Specifies the kernel type to be used in the algorithm.
     It must be one of 'linear', 'poly', 'rbf', 'sigmoid', 'precomputed' or
     a callable.
     If none is given, 'rbf' will be used. If a callable is given it is
     used to precompute the kernel matrix.

nu : float, optional
    An upper bound on the fraction of training
    errors and a lower bound of the fraction of support
    vectors. Should be in the interval (0, 1]. By default 0.5
    will be taken.

degree : int, optional (default=3)
    Degree of the polynomial kernel function ('poly').
    Ignored by all other kernels.

gamma : float, optional (default='auto')
    Kernel coefficient for 'rbf', 'poly' and 'sigmoid'.
    If gamma is 'auto' then 1/n_features will be used instead.

coef0 : float, optional (default=0.0)
    Independent term in kernel function.
    It is only significant in 'poly' and 'sigmoid'.

tol : float, optional
    Tolerance for stopping criterion.

shrinking : bool, optional
    Whether to use the shrinking heuristic.

cache_size : float, optional
    Specify the size of the kernel cache (in MB).

verbose : bool, default: False
    Enable verbose output. Note that this setting takes advantage of a
    per-process runtime setting in libsvm that, if enabled, may not work
    properly in a multithreaded context.

max_iter : int, optional (default=-1)
    Hard limit on iterations within solver, or -1 for no limit.

contamination : float in (0., 0.5), optional (default=0.1)
    The amount of contamination of the data set, i.e.
    the proportion of outliers in the data set. Used when fitting to
    define the threshold on the decision function.
"""

Auto = Literal["auto"]
Kernel = Literal["linear", "poly", "rbf", "sigmoid", "precomputed"]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class OCSVMDetectorConfig(PyODModelConfig):
    kernel: Kernel | Callable = "rbf"
    nu: float = 0.5
    degree: int = 3
    gamma: float | Auto = "auto"
    coef0: float = 0
    tol: float = 1e-3
    shrinking: bool = True
    cache_size: float = 200
    verbose: bool = False
    max_iter: int = -1
    contamination: float = 0.1


class OCSVMDetector(PyODClassificationMixin, PyODModelBase[OCSVMDetectorConfig]):
    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[OCSVMDetectorConfig]:
        return OCSVMDetectorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> OCSVM:
        config = self.model_config
        return OCSVM(
            kernel=config.kernel,
            nu=config.nu,
            degree=config.degree,
            gamma=config.gamma,
            coef0=config.coef0,
            tol=config.tol,
            shrinking=config.shrinking,
            cache_size=config.cache_size,
            verbose=config.verbose,
            max_iter=config.max_iter,
            contamination=config.contamination,
        )



from typing import Any

try:
    from rdkit import Chem, rdBase

except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None


class RDKitLogContext:
    """Context manager to disable RDKit logs. By default all logs are disabled."""

    def __init__(
        self,
        mute_errors: bool = True,
        mute_warning: bool = True,
        mute_info: bool = True,
        mute_debug: bool = True,
    ):
        # Get current log state
        self.previous_status = self._get_log_status()

        # Init the desired log state to apply during in the context
        self.desired_status = {}
        self.desired_status["rdApp.error"] = not mute_errors
        self.desired_status["rdApp.warning"] = not mute_warning
        self.desired_status["rdApp.debug"] = not mute_debug
        self.desired_status["rdApp.info"] = not mute_info

    def _get_log_status(self) -> Any:
        """Get the current log status of RDKit logs."""
        log_status = rdBase.LogStatus()
        log_status = {
            st.split(":")[0]: st.split(":")[1] for st in log_status.split("\n")
        }
        log_status = {
            k: True if v == "enabled" else False for k, v in log_status.items()
        }
        return log_status

    def _apply_log_status(self, log_status: Any) -> None:
        """Apply an RDKit log status."""
        for k, v in log_status.items():
            if v is True:
                rdBase.EnableLog(k)
            else:
                rdBase.DisableLog(k)

    def __enter__(self) -> None:
        self._apply_log_status(self.desired_status)

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self._apply_log_status(self.previous_status)


def rdkit_mol_from_smiles(smiles: str) -> Chem.Mol:
    """Returns a Chem.Mol object representing the input SMILES string."""
    try:
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            raise RuntimeError(f"Error parsing {smiles} using RDKit")
        return mol

    except TypeError as e:
        raise TypeError(
            f"{smiles} of type {type(smiles)} not supported by RDKit",
        ) from e


def rdkit_mol_from_bytes(bytes_molecule: bytes) -> Chem.Mol:
    """Returns a Chem.Mol object representing the input bytes string."""
    rdkit_mol = Chem.Mol(bytes_molecule)
    if rdkit_mol:
        return rdkit_mol


def rdkit_mol_from_hex(hex_molecule: str) -> Chem.Mol:
    """Returns a Chem.Mol object representing the input hex string."""
    bytes_molecule = bytes.fromhex(hex_molecule)
    return rdkit_mol_from_bytes(bytes_molecule)


def smiles_from_rdkit_mol(rdkit_mol: Chem.Mol) -> str:
    """
    Returns a SMILES string representing the input Chem.Mol object.
    """
    smiles: str = Chem.MolToSmiles(rdkit_mol)
    return smiles


def smiles_from_bytes(bytes_molecule: bytes) -> str:
    """
    Returns a SMILES string representing the input bytes string.
    """
    rdkit_mol = rdkit_mol_from_bytes(bytes_molecule)
    return smiles_from_rdkit_mol(rdkit_mol)


def to_rdkit_mol(molecule: Any) -> Chem.Mol:
    """
    Converts a single input to an Chem.Mol object. This is done by going over the following cases in order:

    - input is a Chem.Mol, in which case it is returned unchanged
    - input is a string
        - if it can be parsed as a SMILES, it is converted to a Chem.Mol from the SMILES
        - else, if it is a hexadecimal string, it is converted to bytes and then to a Chem.Mol
        - else, raise a TypeError
    - input is a bytes
        - it is converted to an Chem.Mol
    - else, raise an error

    Args:
        molecule: Any

    Returns:
        Chem.Mol
    """
    with RDKitLogContext():
        if isinstance(molecule, Chem.Mol):
            return molecule

        if isinstance(molecule, str):
            try:
                rdkit_mol = rdkit_mol_from_smiles(molecule)
                if rdkit_mol is not None:
                    return rdkit_mol
            except:  # noqa: E722 S110
                pass

            try:
                return rdkit_mol_from_hex(molecule)
            except ValueError:
                pass

            raise TypeError(
                f"Unparsable string input sample: {molecule}. Tried to parse as SMILES and HEX.",
            )

        elif isinstance(molecule, bytes):
            return rdkit_mol_from_bytes(molecule)

        else:
            raise TypeError(f"Unsupported input sample type: {type(molecule)!r}")


def to_smiles(molecule: Any) -> str:
    """
    Converts a single sample to a SMILES string.
    """
    with RDKitLogContext():
        if isinstance(molecule, str):
            mol = Chem.MolFromSmiles(molecule)
            if mol:
                return molecule

            try:
                return smiles_from_rdkit_mol(rdkit_mol_from_hex(molecule))
            except ValueError:
                pass

            raise TypeError(
                f"Unparsable string input sample: {molecule}. Tried to parse as SMILES and HEX.",
            )

        elif isinstance(molecule, Chem.Mol):
            return smiles_from_rdkit_mol(molecule)

        elif isinstance(molecule, bytes):
            return smiles_from_bytes(molecule)

        else:
            raise TypeError(f"Unsupported input sample type: {type(molecule)!r}")


def to_hex(molecule: Any) -> str:
    """
    Converts a single sample to bytes and then to a hex string.
    """

    mol = to_rdkit_mol(molecule)

    mol_bytes = mol.ToBinary(Chem.PropertyPickleOptions.AllProps)

    mol_hex = str(mol_bytes.hex())

    return mol_hex



import pytest

from molflux.splits.presets import train_validation_test_split
from molflux.splits.strategies.core.group_shuffle_split import GroupShuffleSplit
from molflux.splits.strategies.core.linear_split import LinearSplit
from molflux.splits.strategies.core.shuffle_split import ShuffleSplit
from molflux.splits.strategies.core.stratified_shuffle_split import (
    StratifiedShuffleSplit,
)


def test_returns_shuffle_split_by_default():
    """That a ShuffleSplit strategy is returned by default"""
    strategy = train_validation_test_split()
    assert isinstance(strategy, ShuffleSplit)


def test_strategy_generates_train_validation_test_splits():
    """That train, validation, test splits are generated."""
    strategy = train_validation_test_split()
    data = range(100)
    folds = strategy.split(data)
    train_indices, validation_indices, test_indices = next(folds)
    assert len(list(train_indices))
    assert len(list(validation_indices))
    assert len(list(test_indices))


def test_returns_shuffle_split_if_shuffle_and_not_stratified_and_not_grouped():
    """That a ShuffleSplit strategy is returned if 'shuffle' is True,
    'stratified' is False, and 'grouped' is False."""
    strategy = train_validation_test_split(
        shuffle=True,
        stratified=False,
        grouped=False,
    )
    assert isinstance(strategy, ShuffleSplit)


def test_returns_stratified_shuffle_split_if_shuffle_and_stratified_and_not_grouped():
    """That a StratifiedShuffleSplit strategy is returned if 'shuffle' is True', and
    'stratified' is True, and 'grouped' is False."""
    strategy = train_validation_test_split(shuffle=True, stratified=True, grouped=False)
    assert isinstance(strategy, StratifiedShuffleSplit)


def test_returns_group_shuffle_split_if_shuffle_and_not_stratified_and_grouped():
    """That a GroupShuffleSplit strategy is returned if 'shuffle' is True,
    'stratified' is False, and 'grouped' is True."""
    strategy = train_validation_test_split(shuffle=True, stratified=False, grouped=True)
    assert isinstance(strategy, GroupShuffleSplit)


def test_returns_linear_split_if_not_shuffle_and_not_stratified():
    """That a LinearSplit strategy is returned if 'shuffle' is True', and
    'stratified' is False."""
    strategy = train_validation_test_split(
        shuffle=False,
        stratified=False,
        grouped=True,
    )
    assert isinstance(strategy, LinearSplit)
    strategy = train_validation_test_split(
        shuffle=False,
        stratified=False,
        grouped=False,
    )
    assert isinstance(strategy, LinearSplit)


def test_raises_on_not_implemented_combination():
    """That an exception is raised if requesting an unsupported combination of
    'shuffle', 'stratified', and 'grouped' parameters."""
    with pytest.raises(NotImplementedError):
        train_validation_test_split(shuffle=False, stratified=True)



import os
from pathlib import Path
from typing import Any, ClassVar

from tdc.benchmark_group import admet_group

import datasets
from datasets.packaged_modules.csv.csv import Csv, CsvConfig

# from tdc.utils.retrieve_benchmark_names("ADMET_Group")
_ADMET_GROUP_BENCHMARK_NAMES = [
    "caco2_wang",
    "hia_hou",
    "pgp_broccatelli",
    "bioavailability_ma",
    "lipophilicity_astrazeneca",
    "solubility_aqsoldb",
    "bbb_martins",
    "ppbr_az",
    "vdss_lombardo",
    "cyp2d6_veith",
    "cyp3a4_veith",
    "cyp2c9_veith",
    "cyp2d6_substrate_carbonmangels",
    "cyp3a4_substrate_carbonmangels",
    "cyp2c9_substrate_carbonmangels",
    "half_life_obach",
    "clearance_microsome_az",
    "clearance_hepatocyte_az",
    "herg",
    "ames",
    "dili",
    "ld50_zhu",
]
_ASSETS_DIR = ".cache/huggingface/admet_benchmark"

_DESCRIPTION = f"""\
TDC ADMET benchmark group.

The following dataset configurations are available:
{_ADMET_GROUP_BENCHMARK_NAMES}
"""

_HOMEPAGE = "https://tdcommons.ai/"

_CITATION = """\
@misc{huang2021therapeutics,
      title={Therapeutics Data Commons: Machine Learning Datasets and Tasks for Drug Discovery and Development},
      author={Kexin Huang and Tianfan Fu and Wenhao Gao and Yue Zhao and Yusuf Roohani and Jure Leskovec and Connor W. Coley and Cao Xiao and Jimeng Sun and Marinka Zitnik},
      year={2021},
      eprint={2102.09548},
      archivePrefix={arXiv},
      primaryClass={cs.LG}
}
"""


class ADMETBenchmarks(Csv):
    """

    Examples:
        >>> import molflux.datasets
        >>> molflux.datasets.load_dataset("tdc_admet_benchmarks", config_name="caco2_wang", split=None)  # doctest: +SKIP
        DatasetDict({
            train_validation: Dataset({
                features: ['Drug_ID', 'Drug', 'Y'],
                num_rows: 728
            })
            test: Dataset({
                features: ['Drug_ID', 'Drug', 'Y'],
                num_rows: 182
            })
        })
        >>> molflux.datasets.load_dataset("tdc_admet_benchmarks", config_name="caco2_wang", split="test")  # doctest: +SKIP
        Dataset({
            features: ['Drug_ID', 'Drug', 'Y'],
            num_rows: 182
        })
        >>> molflux.datasets.load_dataset("tdc_admet_benchmarks", config_name="caco2_wang", split="all")  # doctest: +SKIP
        Dataset({
            features: ['Drug_ID', 'Drug', 'Y'],
            num_rows: 910
        })
    """

    BUILDER_CONFIGS: ClassVar[list[datasets.BuilderConfig]] = [
        CsvConfig(name=name) for name in _ADMET_GROUP_BENCHMARK_NAMES
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def _split_generators(
        self,
        dl_manager: datasets.DownloadManager,
    ) -> list[datasets.SplitGenerator]:
        # Link up to the appropriate assets subdirectory
        cache_path = Path.home() / Path(_ASSETS_DIR)
        cache_path.mkdir(exist_ok=True)
        admet_group(path=cache_path)

        data_path = os.path.join(str(cache_path / "admet_group"), self.config.name)
        self.config.data_files = {
            "train_validation": os.path.join(data_path, "train_val.csv"),
            "test": os.path.join(data_path, "test.csv"),
        }

        return super()._split_generators(dl_manager=dl_manager)  # type: ignore[no-any-return]



"""
Tests ensuring desired API objects are part of the top-level namespace.
"""

import molflux.metrics


def test_exports_list_metrics():
    """That the package exposes the list_metrics function."""
    assert hasattr(molflux.metrics, "list_metrics")
    assert callable(molflux.metrics.list_metrics)


def test_exports_list_suites():
    """That the package exposes the list_suites function."""
    assert hasattr(molflux.metrics, "list_suites")
    assert callable(molflux.metrics.list_suites)


def test_exports_load_from_dict():
    """That the package exposes the load_from_dict function."""
    assert hasattr(molflux.metrics, "load_from_dict")
    assert callable(molflux.metrics.load_from_dict)


def test_exports_load_from_dicts():
    """That the package exposes the load_from_dicts function."""
    assert hasattr(molflux.metrics, "load_from_dict")
    assert callable(molflux.metrics.load_from_dicts)


def test_exports_load_from_yaml():
    """That the package exposes the load_from_yaml function."""
    assert hasattr(molflux.metrics, "load_from_yaml")
    assert callable(molflux.metrics.load_from_yaml)


def test_exports_load_metric():
    """That the package exposes the load_metric function."""
    assert hasattr(molflux.metrics, "load_metric")
    assert callable(molflux.metrics.load_metric)


def test_exports_load_metrics():
    """That the package exposes the load_metrics function."""
    assert hasattr(molflux.metrics, "load_metrics")
    assert callable(molflux.metrics.load_metrics)


def test_exports_load_suite():
    """That the package exposes the load_suite function."""
    assert hasattr(molflux.metrics, "load_suite")
    assert callable(molflux.metrics.load_suite)


def test_exports_metric():
    """That the package exposes the Metric protocol."""
    assert hasattr(molflux.metrics, "Metric")


def test_exports_metrics():
    """That the package exposes the Metrics class."""
    assert hasattr(molflux.metrics, "Metrics")


def test_exports_supports_prediction_intervals():
    """That the package exposes the supports_uncertainty function."""
    assert hasattr(molflux.metrics, "supports_prediction_intervals")
    assert callable(molflux.metrics.supports_prediction_intervals)



"""
Wrappers for generating commonly used splitting strategies.
"""

from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy


def train_validation_test_split(
    n_splits: int = 1,
    train_fraction: float = 0.8,
    validation_fraction: float = 0.1,
    test_fraction: float = 0.1,
    seed: int | None = None,
    shuffle: bool = True,
    stratified: bool = False,
    grouped: bool = False,
) -> SplittingStrategy:
    if shuffle and not stratified and not grouped:
        strategy_name = "shuffle_split"
    elif shuffle and stratified and not grouped:
        strategy_name = "stratified_shuffle_split"
    elif shuffle and not stratified and grouped:
        strategy_name = "group_shuffle_split"
    elif not shuffle and not stratified:
        strategy_name = "linear_split"
    else:
        raise NotImplementedError("The requested splitting strategy is not available.")

    strategy = load_splitting_strategy(name=strategy_name)

    strategy.update_state(
        n_splits=n_splits,
        train_fraction=train_fraction,
        validation_fraction=validation_fraction,
        test_fraction=test_fraction,
        seed=seed,
    )

    return strategy


def train_test_split(
    n_splits: int = 1,
    train_fraction: float = 0.8,
    seed: int | None = None,
    shuffle: bool = True,
    stratified: bool = False,
    grouped: bool = False,
) -> SplittingStrategy:
    return train_validation_test_split(
        n_splits=n_splits,
        train_fraction=train_fraction,
        validation_fraction=0,
        test_fraction=1 - train_fraction,
        seed=seed,
        shuffle=shuffle,
        stratified=stratified,
        grouped=grouped,
    )


def k_fold_split(
    n_splits: int = 5,
    shuffle: bool = False,
    seed: int | None = None,
    stratified: bool = False,
    grouped: bool = False,
) -> SplittingStrategy:
    if stratified and not grouped:
        strategy_name = "stratified_k_fold"
    elif grouped and not stratified:
        strategy_name = "group_k_fold"
    elif not stratified and not grouped:
        strategy_name = "k_fold"
    else:
        raise NotImplementedError("The requested splitting strategy is not available.")

    strategy = load_splitting_strategy(name=strategy_name)

    strategy.update_state(
        n_splits=n_splits,
        shuffle=shuffle,
        seed=seed,
    )

    return strategy



from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import SKLearnModelBase

try:
    from sklearn.kernel_ridge import KernelRidge as KR
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
Kernel ridge regression (KRR) combines ridge regression (linear least
squares with l2-norm regularization) with the kernel trick. It thus
learns a linear function in the space induced by the respective kernel and
the data. For non-linear kernels, this corresponds to a non-linear
function in the original space.

The form of the model learned by KRR is identical to support vector
regression (SVR). However, different loss functions are used: KRR uses
squared error loss while support vector regression uses epsilon-insensitive
loss, both combined with l2 regularization. In contrast to SVR, fitting a
KRR model can be done in closed-form and is typically faster for
medium-sized datasets. On the other hand, the learned model is non-sparse
and thus slower than SVR, which learns a sparse model for epsilon > 0, at
prediction-time.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
alpha : float or array-like of shape (n_targets,), default=1.0
    Regularization strength; must be a positive float. Regularization
    improves the conditioning of the problem and reduces the variance of
    the estimates. Larger values specify stronger regularization.
    Alpha corresponds to ``1 / (2C)`` in other linear models such as
    :class:`~sklearn.linear_model.LogisticRegression` or
    :class:`~sklearn.svm.LinearSVC`. If an array is passed, penalties are
    assumed to be specific to the targets. Hence they must correspond in
    number.
kernel : str or callable, default="linear"
    Kernel mapping used internally. This parameter is directly passed to
    :class:`~sklearn.metrics.pairwise.pairwise_kernel`.
    If `kernel` is a string, it must be one of the metrics
    in `pairwise.PAIRWISE_KERNEL_FUNCTIONS` or "precomputed".
    If `kernel` is "precomputed", X is assumed to be a kernel matrix.
    Alternatively, if `kernel` is a callable function, it is called on
    each pair of instances (rows) and the resulting value recorded. The
    callable should take two rows from X as input and return the
    corresponding kernel value as a single number. This means that
    callables from :mod:`sklearn.metrics.pairwise` are not allowed, as
    they operate on matrices, not single samples. Use the string
    identifying the kernel instead.
gamma : float, default=None
    Gamma parameter for the RBF, laplacian, polynomial, exponential chi2
    and sigmoid kernels. Interpretation of the default value is left to
    the kernel; see the documentation for sklearn.metrics.pairwise.
    Ignored by other kernels.
degree : float, default=3
    Degree of the polynomial kernel. Ignored by other kernels.
coef0 : float, default=1
    Zero coefficient for polynomial and sigmoid kernels.
    Ignored by other kernels.
kernel_params : mapping of str to any, default=None
    Additional parameters (keyword arguments) for kernel function passed
    as callable object.
"""


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class KernelRidgeConfig(ModelConfig):
    alpha: float | list[float] = 1.0
    kernel: str = "linear"
    gamma: float | None = None
    degree: int = 3
    coef0: float = 1.0
    kernel_params: dict | None = None


class KernelRidgeRegressor(SKLearnModelBase[KernelRidgeConfig]):
    @property
    def _config_builder(self) -> type[KernelRidgeConfig]:
        return KernelRidgeConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> KR:
        config = self.model_config
        return KR(
            alpha=config.alpha,
            kernel=config.kernel,
            gamma=config.gamma,
            degree=config.degree,
            coef0=config.coef0,
            kernel_params=config.kernel_params,
        )



import itertools

import numpy as np
import pytest

from molflux.splits.catalogue import list_splitting_strategies
from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy

strategy_name = "stratified_shuffle_split"


@pytest.fixture(scope="module")
def fixture_test_strategy():
    return load_splitting_strategy(strategy_name)


@pytest.fixture(scope="module")
def fixture_sample_inputs():
    n = 100
    dataset = np.random.rand(n)
    classes = np.random.choice([0, 1, 2], size=n)
    return dataset, classes


def test_is_in_catalogue():
    """That the strategy is registered in the catalogue."""
    catalogue = list_splitting_strategies()
    all_strategy_names = [name for names in catalogue.values() for name in names]
    assert strategy_name in all_strategy_names


def test_implements_protocol(fixture_test_strategy):
    """That the strategy implements the protocol."""
    strategy = fixture_test_strategy
    assert isinstance(strategy, SplittingStrategy)


def test_requires_classes(fixture_test_strategy):
    """That classes need to be provided for stratification strategy."""
    dataset = [0, 1, 2, 3, 4, 5]
    strategy = fixture_test_strategy
    with pytest.raises(ValueError, match="The 'y' parameter should not be None"):
        next(strategy.split(dataset))


def test_accepts_integer_class_labels(fixture_test_strategy):
    """That the splitting strategy accepts integer class labels."""
    n = 100
    dataset = np.random.rand(n)
    classes = np.random.choice([0, 1, 2], size=n)
    strategy = fixture_test_strategy
    next(strategy.split(dataset, y=classes))
    assert True


def test_accepts_string_class_labels(fixture_test_strategy):
    """That the splitting strategy accepts string class labels."""
    n = 100
    dataset = np.random.rand(n)
    classes = np.random.choice(["cat", "dog", "snake"], size=n)
    strategy = fixture_test_strategy
    next(strategy.split(dataset, y=classes))
    assert True


def test_yields_one_fold_by_default(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy only yields one set of splits by default."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, y=classes)
    assert len(list(indices)) == 1


def test_can_yield_multiple_folds(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy can generate multiple folds."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, y=classes, n_splits=2)
    assert len(list(indices)) == 2


def test_non_deterministic_split(fixture_sample_inputs, fixture_test_strategy):
    """That split results are non deterministic."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy

    indices = strategy.split(dataset, y=classes, n_splits=2)
    train_indices1, validation_indices1, test_indices1 = next(indices)
    train_indices2, validation_indices2, test_indices2 = next(indices)

    assert any(i != j for i, j in zip(train_indices1, train_indices2, strict=False))
    assert any(
        i != j for i, j in zip(validation_indices1, validation_indices2, strict=False)
    )
    assert any(i != j for i, j in zip(test_indices1, test_indices2, strict=False))


def test_deterministic_split_if_seed_set(fixture_sample_inputs, fixture_test_strategy):
    """That split results are deterministic if a seed is set."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy

    indices1 = strategy.split(dataset, y=classes, seed=123)
    train_indices1, validation_indices1, test_indices1 = next(indices1)

    indices2 = strategy.split(dataset, y=classes, seed=123)
    train_indices2, validation_indices2, test_indices2 = next(indices2)

    assert all(i == j for i, j in zip(train_indices1, train_indices2, strict=False))
    assert all(
        i == j for i, j in zip(validation_indices1, validation_indices2, strict=False)
    )
    assert all(i == j for i, j in zip(test_indices1, test_indices2, strict=False))


def test_splits_are_shuffled_across_folds(fixture_sample_inputs, fixture_test_strategy):
    """That even if a seed is set, output folds are individually shuffled.

    This is to align with the expected behaviour in the scikit-learn API.
    """
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy

    indices1 = strategy.split(dataset=dataset, y=classes, n_splits=2, seed=123)
    train_indices1, validation_indices1, test_indices1 = next(indices1)
    train_indices2, validation_indices2, test_indices2 = next(indices1)

    assert any(i != j for i, j in zip(train_indices1, train_indices2, strict=False))
    assert any(
        i != j for i, j in zip(validation_indices1, validation_indices2, strict=False)
    )
    assert any(i != j for i, j in zip(test_indices1, test_indices2, strict=False))


def test_splits_are_disjoint(fixture_sample_inputs, fixture_test_strategy):
    """That data is spread across splits without overlap."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, y=classes)
    train_indices, validation_indices, test_indices = next(indices)
    assert set(train_indices).isdisjoint(set(validation_indices))
    assert set(train_indices).isdisjoint(set(test_indices))
    assert set(validation_indices).isdisjoint(set(test_indices))


def test_inconsistent_split_fractions_raise(
    fixture_sample_inputs,
    fixture_test_strategy,
):
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(
        dataset=dataset,
        y=classes,
        train_fraction=0.6,
        validation_fraction=0.1,
        test_fraction=0.1,
    )
    with pytest.raises(AssertionError):
        next(indices)


def test_not_missing_some_indices(fixture_test_strategy):
    """
    For small datasets and specific conditions, the output splits were missing
    some indices. We test against a specific example here to catch deviations
    from a known state, but we don't rule out that the exact results may
    change over time (e.g. for changes in seeding algorithm)
    """

    dataset = range(10)
    classes = ["a", "a", "a", "b", "b", "b", "b", "b", "b", "b"]

    strategy = fixture_test_strategy
    indices = strategy.split(
        n_splits=1,
        dataset=dataset,
        y=classes,
        train_fraction=0.8,
        validation_fraction=0.0,
        test_fraction=0.2,
        seed=1,
    )
    train_indices, validation_indices, test_indices = next(indices)

    assert len(
        list(train_indices) + list(validation_indices) + list(test_indices),
    ) == len(dataset)

    assert list(train_indices) == [8, 4, 6, 5, 9, 2, 3, 0]
    assert list(validation_indices) == []
    assert list(test_indices) == [7, 1]


def test_not_always_the_same_indices_in_each_split(fixture_test_strategy):
    """That the class indices dispatched to each split are being shuffled across
     folds.

    We don't want something like this happening:

        [8, 3, 1, 7, 0, 5, 4, 6] [] [2, 9]
        [3, 4, 5, 6, 1, 7, 0, 8] [] [9, 2]
        [8, 6, 4, 0, 7, 5, 3, 1] [] [9, 2]
        [0, 8, 1, 4, 5, 3, 7, 6] [] [2, 9]
        [3, 8, 4, 5, 1, 6, 0, 7] [] [2, 9]
        ...

    but rather something like this, where it's not always the same class
    indices being dispatched to each split:

        [8 4 6 5 9 2 3 0] [] [7 1]
        [2 5 0 9 8 4 3 6] [] [1 7]
        [3 5 7 4 6 1 8 0] [] [2 9]
        [7 6 4 0 8 9 5 1] [] [2 3]
        [5 8 6 4 2 0 9 7] [] [1 3]
        ...

    For large enough number of iterations, all indices should end up having been
    seen each at least once in all (non-empty) splits.
    """

    dataset = range(10)
    classes = ["a", "a", "a", "b", "b", "b", "b", "b", "b", "b"]

    # pick a large enough number to make all indices show up in all splits
    # over time
    n_splits = 20

    strategy = fixture_test_strategy
    indices = strategy.split(
        n_splits=n_splits,
        dataset=dataset,
        y=classes,
        train_fraction=0.8,
        validation_fraction=0.0,
        test_fraction=0.2,
        seed=1,
    )

    all_test_indices = set(itertools.chain.from_iterable(t for _, _, t in indices))
    for idx in range(len(dataset)):
        assert idx in all_test_indices



from typing import Any

try:
    from openeye import oegraphsim
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import (
    fingerprint_to_bit_vector,
    to_oemol,
)
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
MACCS fingerprint. [OpenEye]

MACCS keys are 166 bit structural key descriptors in which each bit is
associated with a SMARTS pattern.
"""


class MACCS(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates MACCS fingerprints for each input molecule.

        Args:
            samples: The molecules to featurise

        Returns:
            MACCS fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('maccs')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples=samples)
            {'maccs': [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        maccs_fp_list: list[Fingerprint] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                # patch openeye support for fingerprints of empty SMILES
                if sample == "":
                    _MACCS_FINGERPRINT_LENGTH = 166
                    bit_vector = [0] * _MACCS_FINGERPRINT_LENGTH

                else:
                    mol = to_oemol(sample)
                    fp = oegraphsim.OEFingerPrint()
                    oegraphsim.OEMakeMACCS166FP(fp, mol)
                    bit_vector = fingerprint_to_bit_vector(fp)

                maccs_fp_list.append(bit_vector)

        return {self.tag: maccs_fp_list}



import numpy
import pytest

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.classification.top_k_accuracy_roc import TopKAccuracyRoc

metric_name = "top_k_accuracy_roc"


@pytest.fixture()
def fixture_metric():
    return load_metric(metric_name)


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert metric_name in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, TopKAccuracyRoc)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_string_labels(fixture_metric):
    """That can use string labels."""
    metric = fixture_metric
    predictions = [["A", "B", "C", "D", "E"]]
    references = ["B"]
    result = metric.compute(predictions=predictions, references=references)
    numpy.testing.assert_allclose([[0.0, 1.0, 1.0, 1.0, 1.0]], result[metric_name])


def test_none_labels(fixture_metric):
    """That can use None labels."""
    metric = fixture_metric
    predictions = [["A", "B", "B", None, "C"]]
    references = ["B"]
    result = metric.compute(predictions=predictions, references=references)
    numpy.testing.assert_allclose([[0.0, 1.0, 1.0, 1.0, 1.0]], result[metric_name])


def test_with_average(fixture_metric):
    """That switching on the average flag returns expected results."""
    metric = fixture_metric
    predictions = [["A", "B", "C", "D", "E"], ["A", "B", "C", "D", "E"]]
    references = ["B", "C"]
    result = metric.compute(
        predictions=predictions,
        references=references,
        average=True,
    )
    numpy.testing.assert_allclose([[0.0, 0.5, 1.0, 1.0, 1.0]], result[metric_name])



import logging
from typing import Any

try:
    import torch
    from class_resolver import ClassResolver
    from lightning.pytorch import LightningModule, Trainer, callbacks
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning", e) from e


logger = logging.getLogger(__name__)

callback_resolver = ClassResolver.from_subclasses(callbacks.Callback)


# You may register custom callbacks too


class ModelCheckpointApply(callbacks.ModelCheckpoint):
    """Model checkpointing callback which applies the best module checkpoint at the end of fitting"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def on_fit_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        del trainer

        if self.best_model_path is not None:
            best_ckpt = torch.load(
                self.best_model_path,
                map_location="cpu",
                weights_only=True,
            )
            pl_module.load_state_dict(best_ckpt["state_dict"])
            logger.warning(
                f"Using best model from ckpt_path: {self.best_model_path}.\n",
            )


callback_resolver.register(ModelCheckpointApply)



"""
Wrappers for using splitting strategies to split datasets.Datasets.
"""

from collections.abc import Iterator

from datasets import Dataset, DatasetDict
from molflux.datasets.interfaces import SplittingStrategy


def split_dataset(
    dataset: Dataset,
    strategy: SplittingStrategy,
    target_column: str | None = None,
    groups_column: str | None = None,
) -> Iterator[DatasetDict]:
    """Generates dataset splits according to given strategy.

    Args:
        dataset: The dataset to split.
        strategy: A pre-configured instance of the splitting strategy to use.
        target_column: The name of the dataset column to use as target variable for
            supervised learning problems ('y') when splitting.
        groups_column: The name of the dataset column to use as group labels when
            splitting.

    Returns:
        A generator of split datasets (folds).
    """

    y = dataset[target_column] if target_column is not None else None
    groups = dataset[groups_column] if groups_column is not None else None

    for train_indices, validation_indices, test_indices in strategy.split(
        dataset=dataset,
        y=y,
        groups=groups,
    ):
        train_split = dataset.select(indices=train_indices)
        validation_split = dataset.select(indices=validation_indices)
        test_split = dataset.select(indices=test_indices)

        yield DatasetDict(
            {
                "train": train_split,
                "validation": validation_split,
                "test": test_split,
            },
        )



import pandas as pd
import pytest
import torch

import datasets
from molflux.modelzoo.load import load_model
from molflux.modelzoo.protocols import Model

model_name = "lightning_mlp_regressor"

_X_FEATURES = ["X_col_1", "X_col_2"]
_Y_FEATURES = ["y_col"]

train_df = pd.DataFrame(
    [
        [1.0, 2.0, 10.0],
        [1.0, 3.0, 11.0],
        [2.0, 4.0, 13.0],
        [2.0, 4.0, 13.0],
        [2.0, 4.0, 13.0],
    ],
    columns=_X_FEATURES + _Y_FEATURES,
)

validation_df = pd.DataFrame(
    [
        [1.0, 2.0, 12.0],
        [1.0, 3.0, 13.0],
        [2.0, 4.0, 14.0],
        [2.0, 4.0, 15.0],
        [2.0, 5.0, 16.0],
    ],
    columns=_X_FEATURES + _Y_FEATURES,
)


@pytest.fixture(scope="function")
def fixture_model() -> Model:
    return load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        input_dim=2,
    )


@pytest.mark.parametrize(
    "train_data, validation_data",
    [
        (
            datasets.Dataset.from_pandas(train_df),
            datasets.Dataset.from_pandas(validation_df),
        ),
    ],
)
def test_model_checkpoint_apply_callback(
    fixture_model,
    tmp_path,
    train_data,
    validation_data,
):
    """Test that ModelCheckpointApply callback applies best ckpt at the end of fit"""

    model = fixture_model

    model.train(
        train_data=train_data,
        validation_data=validation_data,
        trainer_config={
            "accelerator": "cpu",
            "max_epochs": 20,
            "default_root_dir": str(tmp_path),
            "enable_checkpointing": True,
            "callbacks": [
                {
                    "name": "ModelCheckpointApply",
                    "config": {
                        "monitor": "val/total/loss",
                        "save_last": True,
                        "save_top_k": 1,
                    },
                },
            ],
        },
        datamodule_config={
            "train": {"batch_size": 2},
            "validation": {"batch_size": 2},
        },
    )

    # iterate through logdir and find best ckpt path. This is found under checkpoints and
    # starts with epoch={num_best_epoch}
    for p in (tmp_path / "lightning_logs" / "version_0" / "checkpoints").iterdir():
        if "epoch=" in str(p):
            best_ckpt_path = str(p)

    best_ckpt = torch.load(best_ckpt_path, weights_only=True)

    # assert that state dicts match
    for k, v in model.module.state_dict().items():
        assert torch.equal(v, best_ckpt["state_dict"][k])



import functools
import inspect
import logging
from collections import defaultdict
from collections.abc import Callable
from importlib.metadata import EntryPoint, entry_points

from molflux.splits.naming import camelcase_to_snakecase
from molflux.splits.strategy import SplittingStrategy

logger = logging.getLogger(__name__)

SplittingStrategyT = type[SplittingStrategy]

NAMESPACE = "molflux.splits.plugins."

# This is where entrypoints will be registered {<name>: <Entrypoint>}
SPLITTING_STRATEGIES_CATALOGUE: dict[str, EntryPoint] = {}


@functools.lru_cache
def fill_catalogue() -> None:
    """Fills the catalogue with entrypoints.

    Only ever call this once for performance reasons.
    """
    entrypoints = [
        entrypoint
        for namespace, entrypoints in entry_points().items()
        if namespace.startswith(NAMESPACE)
        for entrypoint in entrypoints
    ]
    for entrypoint in entrypoints:
        put_splitting_strategy_entrypoint(splitting_strategy_entrypoint=entrypoint)


def put_splitting_strategy_entrypoint(
    splitting_strategy_entrypoint: EntryPoint,
) -> None:
    """Puts a splitting strategy entrypoint in the catalogue."""
    splitting_strategy_name = camelcase_to_snakecase(splitting_strategy_entrypoint.name)

    if splitting_strategy_name in SPLITTING_STRATEGIES_CATALOGUE:
        if (
            splitting_strategy_entrypoint
            == SPLITTING_STRATEGIES_CATALOGUE[splitting_strategy_name]
        ):
            pass
        else:
            raise KeyError(f"Duplicate splitting strategy: {splitting_strategy_name!r}")

    SPLITTING_STRATEGIES_CATALOGUE[splitting_strategy_name] = (
        splitting_strategy_entrypoint
    )


def get_splitting_strategy_entrypoint(splitting_strategy_name: str) -> EntryPoint:
    """Returns a specific splitting strategy entrypoint from the catalogue.

    The entrypoint is keyed only by name. This allows to decouple from
    eventual reshuffles of entrypoints across groups.
    """
    splitting_strategy_entrypoint = SPLITTING_STRATEGIES_CATALOGUE.get(
        splitting_strategy_name,
    )
    if splitting_strategy_entrypoint is None:
        raise NotImplementedError(
            f"Splitting strategy {splitting_strategy_name!r} is not available.",
        )
    return splitting_strategy_entrypoint


def get_splitting_strategy_cls(splitting_strategy_name: str) -> SplittingStrategyT:
    entrypoint = get_splitting_strategy_entrypoint(splitting_strategy_name)

    # Might raise (e.g. missing dependency extras)
    splitting_strategy_cls = entrypoint.load()

    if not isinstance(splitting_strategy_cls, SplittingStrategy):
        raise NotImplementedError(
            f"Plugin {entrypoint.value} does not implement SplittingStrategy protocol.",
        )

    return splitting_strategy_cls  # type: ignore[return-value]


def list_splitting_strategies() -> dict[str, list[str]]:
    """List all available splitting strategies.

    The catalogue is returned as a view of strategy names keyed by kind.
    """

    view: dict[str, list[str]] = defaultdict(list)

    for (
        splitting_strategy_name,
        splitting_strategy_entrypoint,
    ) in SPLITTING_STRATEGIES_CATALOGUE.items():
        splitting_strategy_kind = splitting_strategy_entrypoint.group.split(".")[-1]
        view[splitting_strategy_kind].append(splitting_strategy_name)

    return dict(sorted(view.items()))


def register_splitting_strategy(kind: str, name: str) -> Callable:
    """Registers a custom splitting strategy in the entrypoints catalogue.

    Examples:
        .. code-block:: python

           @register_splitting_strategy(kind="custom", name="my_strategy")
           class CustomSplittingStrategy:
               ...
    """

    def wrapper(splitting_strategy_cls: SplittingStrategyT) -> SplittingStrategyT:
        module = inspect.getmodule(splitting_strategy_cls).__name__  # type: ignore
        value = f"{module}:{splitting_strategy_cls.__name__}"
        group = NAMESPACE + kind
        entrypoint = EntryPoint(name=name, value=value, group=group)
        put_splitting_strategy_entrypoint(splitting_strategy_entrypoint=entrypoint)
        return splitting_strategy_cls

    return wrapper



"""Vendored with minor modifications from DRFP

https://raw.githubusercontent.com/reymond-group/drfp/main/src/drfp/fingerprint.py
https://github.com/reymond-group/drfp/commit/6c5db5cc7d2057e932cb1e1be8d697dc2c3327e1
"""

from collections.abc import Iterable
from hashlib import blake2b
from typing import Any

import numpy as np
from rdkit.Chem import AllChem
from rdkit.Chem.rdchem import Mol


class NoReactionError(Exception):
    """Raised when the encoder attempts to encode a non-reaction SMILES.

    Attributes:
        message: a message containing the non-reaction SMILES
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class DrfpEncoder:
    """A class for encoding SMILES as drfp fingerprints."""

    @staticmethod
    def shingling_from_mol(
        in_mol: Mol,
        max_radius: int = 3,
        include_rings: bool = True,
        min_radius: int = 0,
        root_central_atom: bool = True,
        include_hydrogens: bool = False,
    ) -> list[bytes]:
        """Creates a molecular shingling from a RDKit molecule (rdkit.Chem.rdchem.Mol).

        Arguments:
            in_mol: A RDKit molecule instance
            max_radius: The drfp radius (a radius of 3 corresponds to drfp6)
            include_rings: Whether or not to include rings in the shingling
            min_radius: The minimum radius that is used to extract n-grams

        Returns:
            The molecular shingling.
        """

        if include_hydrogens:
            in_mol = AllChem.AddHs(in_mol)

        shingling = []

        if include_rings:
            for ring in AllChem.GetSymmSSSR(in_mol):
                bonds = set()
                ring = list(ring)
                indices = set()
                for i in ring:
                    for j in ring:
                        if i != j:
                            indices.add(i)
                            indices.add(j)
                            bond = in_mol.GetBondBetweenAtoms(i, j)
                            if bond is not None:
                                bonds.add(bond.GetIdx())

                ngram = AllChem.MolToSmiles(
                    AllChem.PathToSubmol(in_mol, list(bonds)),
                    canonical=True,
                    allHsExplicit=True,
                ).encode("utf-8")

                shingling.append(ngram)

        if min_radius == 0:
            for atom in in_mol.GetAtoms():
                ngram = atom.GetSmarts().encode("utf-8")
                shingling.append(ngram)

        for index, _ in enumerate(in_mol.GetAtoms()):
            for i in range(1, max_radius + 1):
                p = AllChem.FindAtomEnvironmentOfRadiusN(
                    in_mol,
                    i,
                    index,
                    useHs=include_hydrogens,
                )
                amap: dict[int, Any] = {}
                submol = AllChem.PathToSubmol(in_mol, p, atomMap=amap)

                if index not in amap:
                    continue

                smiles = ""

                if root_central_atom:
                    smiles = AllChem.MolToSmiles(
                        submol,
                        rootedAtAtom=amap[index],
                        canonical=True,
                        allHsExplicit=True,
                    )
                else:
                    smiles = AllChem.MolToSmiles(
                        submol,
                        canonical=True,
                        allHsExplicit=True,
                    )

                if smiles != "":
                    shingling.append(smiles.encode("utf-8"))

        # Set ensures that the same shingle is not hashed multiple times
        # (which would not change the hash, since there would be no new minima)
        return list(set(shingling))

    @staticmethod
    def internal_encode(
        in_smiles: str,
        max_radius: int = 3,
        min_radius: int = 0,
        include_rings: bool = True,
        root_central_atom: bool = True,
        include_hydrogens: bool = False,
    ) -> tuple[np.ndarray, list[bytes]]:
        """Creates an drfp array from a reaction SMILES string.

        Arguments:
            in_smiles: A valid reaction SMILES string
            max_radius: The drfp radius (a radius of 3 corresponds to drfp6)
            min_radius: The minimum radius that is used to extract n-grams
            include_rings: Whether or not to include rings in the shingling

        Returns:
            A tuple with two arrays, the first containing the drfp hash values, the second the substructure SMILES
        """

        sides = in_smiles.split(">")
        if len(sides) < 3:
            raise NoReactionError(
                f"The following is not a valid reaction SMILES: '{in_smiles}'",
            )

        if len(sides[1]) > 0:
            sides[0] += "." + sides[1]

        left = sides[0].split(".")
        right = sides[2].split(".")

        left_shingles: set[bytes] = set()
        right_shingles: set[bytes] = set()

        for component in left:
            mol = AllChem.MolFromSmiles(component)

            sh = DrfpEncoder.shingling_from_mol(
                mol,
                max_radius=max_radius,
                include_rings=include_rings,
                min_radius=min_radius,
                root_central_atom=root_central_atom,
                include_hydrogens=include_hydrogens,
            )

            for s in sh:
                right_shingles.add(s)

        for component in right:
            mol = AllChem.MolFromSmiles(component)

            sh = DrfpEncoder.shingling_from_mol(
                mol,
                max_radius=max_radius,
                include_rings=include_rings,
                min_radius=min_radius,
                root_central_atom=root_central_atom,
                include_hydrogens=include_hydrogens,
            )

            for s in sh:
                left_shingles.add(s)

        s_diff = right_shingles.symmetric_difference(left_shingles)

        return DrfpEncoder.hash(list(s_diff)), list(s_diff)

    @staticmethod
    def hash(shingling: list[bytes]) -> np.ndarray:
        """Directly hash all the SMILES in a shingling to a 32-bit integerself.

        Arguments:
            shingling: A list of n-grams

        Returns:
            A list of hashed n-grams
        """

        return np.array(
            [int(blake2b(t, digest_size=4).hexdigest(), 16) for t in shingling],
        ).astype(np.int32)

    @staticmethod
    def fold(
        hash_values: np.ndarray,
        length: int = 2048,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Folds the hash values to a binary vector of a given length.

        Arguments:
            hash_value: An array containing the hash values
            length: The length of the folded fingerprint

        Returns:
            A tuple containing the folded fingerprint and the indices of the on bits
        """

        folded = np.zeros(length, dtype=np.uint8)
        on_bits = hash_values % length
        folded[on_bits] = 1

        return folded, on_bits

    @staticmethod
    def encode(
        reactions: Iterable[str],
        length: int = 2048,
        min_radius: int = 0,
        max_radius: int = 3,
        include_rings: bool = True,
        root_central_atom: bool = True,
        include_hydrogens: bool = False,
    ) -> list[list[int]]:
        """Encodes a list of reaction SMILES using the drfp fingerprint.

        Args:
            reactions: An iterable (e.g. list) of reaction SMILES to be encoded
            length: The folded length of the fingerprint (the parameter for the modulo hashing)
            min_radius: The minimum radius of a substructure (0 includes single atoms)
            max_radius: The maximum radius of a substructure
            include_rings: Whether to include full rings as substructures
            root_central_atom: Whether to root the central atom of substructures when generating SMILES

        Returns:
            A list of drfp fingerprints.
        """

        return [
            DrfpEncoder.encode_one(
                reaction,
                length=length,
                min_radius=min_radius,
                max_radius=max_radius,
                include_rings=include_rings,
                root_central_atom=root_central_atom,
                include_hydrogens=include_hydrogens,
            )
            for reaction in reactions
        ]

    @staticmethod
    def encode_one(
        reaction: str,
        length: int,
        min_radius: int,
        max_radius: int,
        include_rings: bool,
        root_central_atom: bool,
        include_hydrogens: bool,
    ) -> list[int]:
        """Encodes a reaction SMILES using the drfp fingerprint.

        Args:
            reaction: A reaction SMILES to be encoded
            length: The folded length of the fingerprint (the parameter for the modulo hashing)
            min_radius: The minimum radius of a substructure (0 includes single atoms)
            max_radius: The maximum radius of a substructure
            include_rings: Whether to include full rings as substructures
            root_central_atom: Whether to root the central atom of substructures when generating SMILES

        Returns:
            A drfp fingerprints.
        """

        hashed_diff, smiles_diff = DrfpEncoder.internal_encode(
            reaction,
            min_radius=min_radius,
            max_radius=max_radius,
            include_rings=include_rings,
            root_central_atom=root_central_atom,
            include_hydrogens=include_hydrogens,
        )

        difference_folded, on_bits = DrfpEncoder.fold(
            hashed_diff,
            length=length,
        )

        return difference_folded.tolist()  # type: ignore[no-any-return]



import pytest
from openeye import oechem


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """

    if not oechem.OEChemIsLicensed("python"):
        pytest.exit("Could not find a valid openeye license!", 1)



import os
import sys
from typing import Any, Literal, Protocol, TypedDict

import datasets

try:
    import lightning.pytorch as pl
    from lightning.pytorch.utilities.combined_loader import CombinedLoader
    from lightning.pytorch.utilities.types import EVAL_DATALOADERS, TRAIN_DATALOADERS
    from torch.utils.data import DataLoader
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning", e) from e

from molflux.modelzoo.models.lightning.config import (
    DataModuleConfig,
    LightningConfig,
    SplitConfig,
)


class DataCollator(Protocol):
    def __call__(self, examples: list[dict[str, Any]]) -> dict[str, Any]: ...


class Datasets(TypedDict):
    train: dict[str | None, datasets.Dataset] | None
    validation: dict[str | None, datasets.Dataset] | None
    test: dict[str | None, datasets.Dataset] | None
    predict: datasets.Dataset | None


class LightningDataModule(pl.LightningDataModule):
    _config_builder = DataModuleConfig

    def __init__(
        self,
        model_config: LightningConfig,
        train_data: dict[str | None, datasets.Dataset] | None = None,
        validation_data: dict[str | None, datasets.Dataset] | None = None,
        test_data: dict[str | None, datasets.Dataset] | None = None,
        predict_data: datasets.Dataset | None = None,
        **kwargs: Any,
    ):
        super().__init__()

        self.model_config = model_config

        self.datasets = self.prepare_datasets(
            train_data=train_data,
            validation_data=validation_data,
            test_data=test_data,
            predict_data=predict_data,
            **kwargs,
        )

    def prepare_dataset(
        self,
        data: datasets.Dataset,
        split: str,
        name: str | None = None,
        **kwargs: Any,
    ) -> datasets.Dataset:
        """A method to prepare any input dataset type into a torch dataset ready for torch dataloaders.

        Some custom models will need to overwrite this method with their own implementation.
        """
        del kwargs

        if split == "predict":
            data = data.with_format("torch", columns=self.model_config.x_features)
        else:
            data = data.with_format(
                "torch",
                columns=self.model_config.resolve_train_features(name),
            )

        return data

    def _get_one_train_dataloader(
        self,
        dataset: datasets.Dataset,
        batch_size: int,
    ) -> DataLoader:
        """Returns one dataloader for training data.

        Implementations should override this method if they require a custom
        sampler or dataloader."""

        return DataLoader(
            dataset,  # pyright: ignore
            batch_size=batch_size,
            shuffle=True,
            pin_memory=True,
            num_workers=self.num_workers,
            persistent_workers=bool(self.num_workers),
            drop_last=self.model_config.datamodule.train.drop_last,
            collate_fn=self.collate_fn,
        )

    def _get_one_eval_dataloader(
        self,
        dataset: datasets.Dataset,
        batch_size: int,
    ) -> DataLoader:
        """Returns one dataloader for validation, test or predict (inference) data.

        Implementations should override this method if they require a custom
        sampler or dataloader."""
        return DataLoader(
            dataset,  # pyright: ignore
            batch_size=batch_size,
            shuffle=False,
            pin_memory=True,
            num_workers=self.num_workers,
            persistent_workers=bool(self.num_workers),
            drop_last=False,
            collate_fn=self.collate_fn,
        )

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """Return dataloaders with training data.

        Overrides an abstract method in PyTorch Lightning. Implementations
        probably don't need to modify this method; look at
        _get_one_train_dataloader instead."""
        datasets = self.datasets["train"]

        if datasets is None:
            return {}

        return CombinedLoader(
            {
                name: self._get_one_train_dataloader(
                    dataset,
                    self._get_batch_size("train", name),
                )
                for name, dataset in datasets.items()
            },
            self.model_config.datamodule.train.mode,
        )

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """Return dataloaders with validation data.

        Overrides an abstract method in PyTorch Lightning. Implementations
        probably don't need to modify this method; look at
        _get_one_eval_dataloader instead."""
        return self._get_eval_dataloaders(
            "validation",
            self.model_config.datamodule.validation.mode,
        )

    def test_dataloader(self) -> EVAL_DATALOADERS:
        """Return dataloaders with test data.

        Overrides an abstract method in PyTorch Lightning. Implementations
        probably don't need to modify this method; look at
        _get_one_eval_dataloader instead."""
        return self._get_eval_dataloaders(
            "test",
            self.model_config.datamodule.test.mode,
        )

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        """Return dataloaders with data for prediction (inference).

        Overrides an abstract method in PyTorch Lightning. Implementations
        probably don't need to modify this method; look at
        _get_one_eval_dataloader instead."""
        # NOTE 11/9/2023 Lightning doesn't support non-sequential predict
        dataset = self.datasets["predict"]

        if dataset is None:
            raise ValueError("Specify a prediction dataset.")

        return self._get_one_eval_dataloader(
            dataset,
            self._get_batch_size("predict", None),
        )

    def _get_eval_dataloaders(
        self,
        split: Literal["validation", "test"],
        multi_data_mode: Literal[
            "max_size_cycle",
            "max_size",
            "min_size",
            "sequential",
        ],
    ) -> EVAL_DATALOADERS:
        datasets = self.datasets[split]

        if datasets is None:
            return {}

        return CombinedLoader(
            {
                name: self._get_one_eval_dataloader(
                    dataset,
                    self._get_batch_size(split, name),
                )
                for name, dataset in datasets.items()
            },
            multi_data_mode,
        )

    def prepare_datasets(
        self,
        train_data: dict[str | None, datasets.Dataset] | None,
        validation_data: dict[str | None, datasets.Dataset] | None,
        test_data: dict[str | None, datasets.Dataset] | None,
        predict_data: datasets.Dataset | None,
        **kwargs: Any,
    ) -> Datasets:
        """Prepares datasets by applying logic in `prepare_dataset` to all
        datasets.

        Most implementations will not have to modify this method."""
        if train_data is not None:
            train_data = {
                k: self.prepare_dataset(v, "train", k, **kwargs)
                for k, v in (train_data or {}).items()
            }

        if validation_data is not None:
            validation_data = {
                k: self.prepare_dataset(v, "validation", k, **kwargs)
                for k, v in (validation_data or {}).items()
            }

        if test_data is not None:
            test_data = {
                k: self.prepare_dataset(v, "test", k, **kwargs)
                for k, v in (test_data or {}).items()
            }

        if predict_data is not None:
            predict_data = self.prepare_dataset(predict_data, "predict", **kwargs)

        return Datasets(
            train=train_data,
            validation=validation_data,
            test=test_data,
            predict=predict_data,
        )

    @property
    def num_workers(self) -> int:
        """Number of workers servicing each individual dataloader."""
        config_num_workers: int | Literal["all"] = (
            self.model_config.datamodule.num_workers
        )

        if isinstance(config_num_workers, int):
            return config_num_workers

        if config_num_workers == "all":
            if sys.platform.startswith("linux"):
                # This is the safer way to check num CPUs on k8s
                return len(os.sched_getaffinity(0))  # pyright: ignore
            return os.cpu_count() or 0  # type: ignore[unreachable,unused-ignore]

        raise ValueError("num_workers should be either an int or 'all'.")

    def _get_batch_size(
        self,
        split: Literal["train", "validation", "test", "predict"],
        dataset_name: str | None,
    ) -> int:
        split_config: SplitConfig = getattr(self.model_config.datamodule, split)

        if isinstance(split_config.batch_size, int):
            return split_config.batch_size
        else:
            if dataset_name is None:
                raise ValueError("Specify the dataset name.")
            return split_config.batch_size[dataset_name]

    @property
    def collate_fn(self) -> DataCollator | None:
        """Collator for dataloaders.

        Derived classes can override this."""
        return None



from typing import Literal

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)

try:
    from sklearn.ensemble import (
        GradientBoostingClassifier as SKGradientBoostingClassifier,
    )
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn gradient boosting classifier model.

This algorithm builds an additive model in a forward stage-wise fashion;
it allows for the optimization of arbitrary differentiable loss functions.
In each stage n_classes_ regression trees are fit on the negative gradient of
the loss function, e.g. binary or multiclass log loss. Binary classification is a
special case where only a single regression tree is induced.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
loss : {'log_loss', 'exponential'}, default='log_loss'
    The loss function to be optimized. 'log_loss' refers to binomial and
    multinomial deviance, the same as used in logistic regression.
    It is a good choice for classification with probabilistic outputs.
    For loss 'exponential', gradient boosting recovers the AdaBoost algorithm.
learning_rate : float, default=0.1
    Learning rate shrinks the contribution of each tree by `learning_rate`.
    There is a trade-off between learning_rate and n_estimators.
    Values must be in the range `[0.0, inf)`.
n_estimators : int, default=100
    The number of boosting stages to perform. Gradient boosting
    is fairly robust to over-fitting so a large number usually
    results in better performance.
    Values must be in the range `[1, inf)`.
subsample : float, default=1.0
    The fraction of samples to be used for fitting the individual base
    learners. If smaller than 1.0 this results in Stochastic Gradient
    Boosting. `subsample` interacts with the parameter `n_estimators`.
    Choosing `subsample < 1.0` leads to a reduction of variance
    and an increase in bias.
    Values must be in the range `(0.0, 1.0]`.
criterion : {'friedman_mse', 'squared_error'}, default='friedman_mse'
    The function to measure the quality of a split. Supported criteria are
    'friedman_mse' for the mean squared error with improvement score by
    Friedman, 'squared_error' for mean squared error. The default value of
    'friedman_mse' is generally the best as it can provide a better
    approximation in some cases.
min_samples_split : int or float, default=2
    The minimum number of samples required to split an internal node:
    - If int, values must be in the range `[2, inf)`.
    - If float, values must be in the range `(0.0, 1.0]` and `min_samples_split`
      will be `ceil(min_samples_split * n_samples)`.
min_samples_leaf : int or float, default=1
    The minimum number of samples required to be at a leaf node.
    A split point at any depth will only be considered if it leaves at
    least ``min_samples_leaf`` training samples in each of the left and
    right branches.  This may have the effect of smoothing the model,
    especially in regression.
    - If int, values must be in the range `[1, inf)`.
    - If float, values must be in the range `(0.0, 1.0)` and `min_samples_leaf`
      will be `ceil(min_samples_leaf * n_samples)`.
min_weight_fraction_leaf : float, default=0.0
    The minimum weighted fraction of the sum total of weights (of all
    the input samples) required to be at a leaf node. Samples have
    equal weight when sample_weight is not provided.
    Values must be in the range `[0.0, 0.5]`.
max_depth : int or None, default=3
    Maximum depth of the individual regression estimators. The maximum
    depth limits the number of nodes in the tree. Tune this parameter
    for best performance; the best value depends on the interaction
    of the input variables. If None, then nodes are expanded until
    all leaves are pure or until all leaves contain less than
    min_samples_split samples.
    If int, values must be in the range `[1, inf)`.
min_impurity_decrease : float, default=0.0
    A node will be split if this split induces a decrease of the impurity
    greater than or equal to this value.
    Values must be in the range `[0.0, inf)`.
    The weighted impurity decrease equation is the following::
        N_t / N * (impurity - N_t_R / N_t * right_impurity
                            - N_t_L / N_t * left_impurity)
    where ``N`` is the total number of samples, ``N_t`` is the number of
    samples at the current node, ``N_t_L`` is the number of samples in the
    left child, and ``N_t_R`` is the number of samples in the right child.
    ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
    if ``sample_weight`` is passed.
init : None or 'zero', default=None
    An estimator object that is used to compute the initial predictions.
    If 'zero', the initial raw predictions are set to zero. By default, a
    ``DummyEstimator`` predicting the classes priors is used.
random_state : int, RandomState instance or None, default=None
    Controls the random seed given to each Tree estimator at each
    boosting iteration.
    In addition, it controls the random permutation of the features at
    each split (see Notes for more details).
    It also controls the random splitting of the training data to obtain a
    validation set if `n_iter_no_change` is not None.
    Pass an int for reproducible output across multiple function calls.
max_features : {'auto', 'sqrt', 'log2'}, int or float, default=None
    The number of features to consider when looking for the best split:
    - If int, values must be in the range `[1, inf)`.
    - If float, values must be in the range `(0.0, 1.0]` and the features
      considered at each split will be `max(1, int(max_features * n_features_in_))`.
    - If 'auto', then `max_features=sqrt(n_features)`.
    - If 'sqrt', then `max_features=sqrt(n_features)`.
    - If 'log2', then `max_features=log2(n_features)`.
    - If None, then `max_features=n_features`.
    Choosing `max_features < n_features` leads to a reduction of variance
    and an increase in bias.
    Note: the search for a split does not stop until at least one
    valid partition of the node samples is found, even if it requires to
    effectively inspect more than ``max_features`` features.
verbose : int, default=0
    Enable verbose output. If 1 then it prints progress and performance
    once in a while (the more trees the lower the frequency). If greater
    than 1 then it prints progress and performance for every tree.
    Values must be in the range `[0, inf)`.
max_leaf_nodes : int, default=None
    Grow trees with ``max_leaf_nodes`` in best-first fashion.
    Best nodes are defined as relative reduction in impurity.
    Values must be in the range `[2, inf)`.
    If `None`, then unlimited number of leaf nodes.
warm_start : bool, default=False
    When set to ``True``, reuse the solution of the previous call to fit
    and add more estimators to the ensemble, otherwise, just erase the
    previous solution.
validation_fraction : float, default=0.1
    The proportion of training data to set aside as validation set for
    early stopping. Values must be in the range `(0.0, 1.0)`.
    Only used if ``n_iter_no_change`` is set to an integer.
n_iter_no_change : int, default=None
    ``n_iter_no_change`` is used to decide if early stopping will be used
    to terminate training when validation score is not improving. By
    default it is set to None to disable early stopping. If set to a
    number, it will set aside ``validation_fraction`` size of the training
    data as validation and terminate training when validation score is not
    improving in all of the previous ``n_iter_no_change`` numbers of
    iterations. The split is stratified.
    Values must be in the range `[1, inf)`.
tol : float, default=1e-4
    Tolerance for the early stopping. When the loss is not improving
    by at least tol for ``n_iter_no_change`` iterations (if set to a
    number), the training stops.
    Values must be in the range `[0.0, inf)`.
ccp_alpha : non-negative float, default=0.0
    Complexity parameter used for Minimal Cost-Complexity Pruning. The
    subtree with the largest cost complexity that is smaller than
    ``ccp_alpha`` will be chosen. By default, no pruning is performed.
    Values must be in the range `[0.0, inf)`.
"""

Loss = Literal["log_loss", "exponential"]
Criterion = Literal["friedman_mse", "squared_error"]
Init = Literal["zero"]
MaxFeatures = Literal["auto", "sqrt", "log2"]


class Config:
    extra = "forbid"
    arbitrary_types_allowed = True
    smart_union = True


@dataclasses.dataclass(config=Config)
class GradientBoostingClassifierConfig(ModelConfig):
    loss: Loss = "log_loss"
    learning_rate: float = 0.1
    n_estimators: int = 100
    subsample: float = 1.0
    criterion: Criterion = "friedman_mse"
    min_samples_split: int | float = 2
    min_samples_leaf: int | float = 1
    min_weight_fraction_leaf: float = 0.0
    max_depth: int | None = 3
    min_impurity_decrease: float = 0.0
    init: Init | None = None
    random_state: int | None = None
    max_features: int | float | MaxFeatures | None = None
    verbose: int = 0
    max_leaf_nodes: int | None = None
    warm_start: bool = False
    validation_fraction: float = 0.1
    n_iter_no_change: int | None = None
    tol: float = 1e-4
    ccp_alpha: float = 0.0

    def __post_init_post_parse__(self) -> None:
        if self.ccp_alpha < 0:
            raise ValueError("Complexity parameter `ccp_alpha` cannot be negative.")


class GradientBoostingClassifier(
    SKLearnClassificationMixin,
    SKLearnModelBase[GradientBoostingClassifierConfig],
):
    @property
    def _config_builder(self) -> type[GradientBoostingClassifierConfig]:
        return GradientBoostingClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKGradientBoostingClassifier:
        config = self.model_config
        return SKGradientBoostingClassifier(
            loss=config.loss,
            learning_rate=config.learning_rate,
            n_estimators=config.n_estimators,
            subsample=config.subsample,
            criterion=config.criterion,
            min_samples_split=config.min_samples_split,
            min_samples_leaf=config.min_samples_leaf,
            min_weight_fraction_leaf=config.min_weight_fraction_leaf,
            max_depth=config.max_depth,
            min_impurity_decrease=config.min_impurity_decrease,
            init=config.init,
            random_state=config.random_state,
            max_features=config.max_features,
            verbose=config.verbose,
            max_leaf_nodes=config.max_leaf_nodes,
            warm_start=config.warm_start,
            validation_fraction=config.validation_fraction,
            n_iter_no_change=config.n_iter_no_change,
            tol=config.tol,
            ccp_alpha=config.ccp_alpha,
        )



from os import PathLike as OSPathLike
from typing import Any, Literal, Union

import datasets

AllowedInferenceMethods = Literal[
    "predict",
    "predict_proba",
    "predict_with_prediction_interval",
    "predict_with_std",
    "sample",
]
PathLike = Union[str, OSPathLike]
Dataset = datasets.Dataset
DatasetType = Union[datasets.Dataset, datasets.DatasetDict]
TasksScores = dict[str, dict[str, Any]]
FoldScores = dict[str, TasksScores]



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.openeye.fingerprints.maccs import MACCS

representation_name = "maccs"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, MACCS)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_fingerprint_length(fixture_representation):
    """That maccs fingerprints have fixed expected length (166)."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
    ]
    result = representation.featurise(samples)
    assert representation_name in result
    fingerprint = result[representation_name][0]
    assert len(fingerprint) == 166


def test_empty_smiles_fingerprint_length(fixture_representation):
    """That empty SMILES strings are processed correctly and give fingerprints of the expected length."""
    representation = fixture_representation
    samples = [""]
    result = representation.featurise(samples)
    fingerprint = result[representation_name][0]
    assert len(fingerprint) == 166


def test_default_compute(fixture_representation):
    """That default featurisation gives expected results."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
    ]
    result = representation.featurise(samples)
    expected_result = [
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            1,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            1,
            1,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            1,
            0,
            1,
            0,
            1,
            0,
            0,
            1,
            0,
            1,
            1,
            0,
        ],
    ]
    assert representation_name in result
    assert result[representation_name] == expected_result


def test_batch_compute(fixture_representation):
    """That batch scoring gives expected results."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
        "CC",
    ]
    result = representation.featurise(samples)
    expected_result = [
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            1,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            1,
            1,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            1,
            0,
            1,
            0,
            1,
            0,
            0,
            1,
            0,
            1,
            1,
            0,
        ],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
    ]
    assert representation_name in result
    assert result[representation_name] == expected_result



# Copyright 2020 The HuggingFace Datasets Authors and the current dataset script contributor.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Accuracy metric."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import accuracy_score

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Accuracy is the proportion of correct predictions among the total number of cases processed. It can be computed with:
Accuracy = (TP + TN) / (TP + TN + FP + FN)
TP: True positive
TN: True negative
FP: False positive
FN: False negative
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated targets as returned by a classifier.
    references: Ground truth (correct) target values.
    normalize: If False, return the number of correctly classified samples.
        Otherwise, return the fraction of correctly classified samples.
    sample_weight: Sample weights.
Returns:
    accuracy: Accuracy score.
Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("accuracy")
    >>> predictions = [0, 1]
    >>> references = [0, 1]
    >>> metric.compute(predictions=predictions, references=references)
    {'accuracy': 1.0}
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class Accuracy(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("int32")),
                    "references": datasets.Sequence(datasets.Value("int32")),
                }
                if self.config_name == "multilabel"
                else {
                    "predictions": datasets.Value("int32"),
                    "references": datasets.Value("int32"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        normalize: bool = True,
        **kwargs: Any,
    ) -> MetricResult:
        score = accuracy_score(
            y_true=references,
            y_pred=predictions,
            normalize=normalize,
            sample_weight=sample_weight,
        )
        return {self.tag: float(score)}



import os
from copy import copy
from typing import Any

import numpy as np
from pydantic.v1 import dataclasses

from datasets import Dataset
from molflux.modelzoo import load_from_store, save_to_store
from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.load import load_from_dict
from molflux.modelzoo.model import (
    ModelBase,
    ModelConfig,
    PredictionIntervalMixin,
    SamplingMixin,
    StandardDeviationMixin,
)
from molflux.modelzoo.models.core.average_features_regressor import (
    AverageFeaturesRegressor,
)
from molflux.modelzoo.models.ensemble._combo.utils import (
    check_parameter,
    get_split_indices,
    list_diff,
)
from molflux.modelzoo.protocols import (
    supports_prediction_interval,
    supports_std,
)
from molflux.modelzoo.typing import PredictionResult
from molflux.modelzoo.utils import format_wrapped_model_tag, pick_features

try:
    from molflux.modelzoo.models.sklearn.linear_regressor import LinearRegressor
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None

_DESCRIPTION = """
Implementation of an ensemble regression model.

An ensemble of models is a collection of models trained separately on the same data that
when combined together may offer better performance than any of the individual models.
Ensemble models often perform well on competition leaderboards, even when ensembling
lots of models that individually may perform poorly.

In this implementation of meta-ensembling (also known as stacking), each of the
base regression models makes a prediction and a meta-regressor is trained
on the predictions of the individual models, optionally including the original
data as additional features.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
base_estimators : list
    A list of modelzoo config dicts to define the models used for training
    the base estimators. An error is raised if the base estimators are not
    provided.
meta_estimator : dict, modelzoo model config, optional (default=None)
    The modelzoo model config for a meta regressor to make the final prediction.
    In the default case, the meta_estimator used depends on ``keep_original``.
    Either a linear regressor is used when the original features are kept, or
    when ``keep_original=False`` then the predictions from the base estimators
    are averaged via the ``average_features_regressor``.
n_folds : int, optional (default=2)
    The number of splits of the training sample.
keep_original : bool, optional (default=False)
    If True, keep the original features for training and predicting.
shuffle_data : bool, optional (default=False)
    If True, shuffle the input data.
random_state : int, RandomState or None, optional (default=None)
    If int, random_state is the seed used by the random
    number generator; If RandomState instance, random_state is the random
    number generator; If None, the random number generator is the
    RandomState instance used by `np.random`.
"""


class Config:
    arbitrary_types_allowed = True


@dataclasses.dataclass(config=Config)
class EnsembleRegressorConfig(ModelConfig):
    base_estimators: list[dict[str, Any]] | None = None
    meta_estimator: dict[str, Any] | None = None  # config for modelzoo regression model
    n_folds: int = 2
    keep_original: bool = False
    shuffle_data: bool = False
    random_state: int | None = None


class EnsembleRegressor(
    PredictionIntervalMixin,
    StandardDeviationMixin,
    SamplingMixin,
    ModelBase[EnsembleRegressorConfig],
):
    """
    Meta ensembling, also known as stacking
    See https://datasciblog.github.io/2016/12/27/a-kagglers-guide-to-model-stacking-in-practice/ for more information.
    See also the `combo` library eg. https://github.com/yzhao062/combo/blob/master/combo/models/classifier_stacking.py
    """

    def __init__(self, tag: str | None = None, **config_kwargs: Any) -> None:
        super().__init__(tag=tag, **config_kwargs)

        config = self.config
        if len(config["base_estimators"]) < 2:
            raise ValueError("At least 2 estimators are required")
        self.base_estimators = [
            load_from_dict(cfg) for cfg in config["base_estimators"]
        ]
        self.n_base_estimators_ = len(self.base_estimators)
        if tag is None:
            self.info.tag = format_wrapped_model_tag(
                self.info.tag,
                [estimator.tag for estimator in self.base_estimators],
            )

        # validate input parameters
        if not isinstance(config["n_folds"], int):
            raise ValueError("n_folds must be an integer variable")
        check_parameter(
            config["n_folds"],
            low=2,
            include_left=True,
            param_name="n_folds",
        )
        self.n_folds = config["n_folds"]

        self.all_models_support_std = all(
            supports_std(model) for model in self.base_estimators
        )
        if config["meta_estimator"] is not None:
            self.meta_estimator = load_from_dict(config["meta_estimator"])
        elif config["keep_original"]:
            meta_estimator_colnames = [
                reg.tag for reg in self.base_estimators
            ] + self.x_features
            self.meta_estimator = LinearRegressor(
                x_features=meta_estimator_colnames,
                y_features=self.y_features,
            )
        elif not config["keep_original"]:
            meta_estimator_colnames = [reg.tag for reg in self.base_estimators]
            meta_estimator_std_colnames = [
                f"{reg.tag}::std" for reg in self.base_estimators
            ]
            # If all base estimators provide uncertainty, use this via a Gaussian mixture.
            # O/W fall back to using std of the estimates themselves
            self.meta_estimator = AverageFeaturesRegressor(
                x_features=meta_estimator_colnames,
                y_features=self.y_features,
                x_std_features=meta_estimator_std_colnames
                if self.all_models_support_std
                else None,
            )

        # set flags
        self.keep_original = config["keep_original"]
        self.shuffle_data = config["shuffle_data"]
        self.random_state = config["random_state"]

    def _config(self) -> EnsembleRegressorConfig:
        return EnsembleRegressorConfig()

    @property
    def _config_builder(self) -> type[EnsembleRegressorConfig]:
        return EnsembleRegressorConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _train(self, train_data: Dataset, **kwargs: Any) -> Any:
        """
        Fit ensemble regressor. Iterate over folds of cross validation set to fit individual ensemble models.
        Fit meta-estimator based on these. Then refit individual models on full data.
        """
        n_samples = train_data.shape[0]

        # initialize matrix for storing newly generated features
        new_features = np.zeros(
            [n_samples, (1 + self.all_models_support_std) * self.n_base_estimators_],
        )

        # build CV datasets
        index_lists = get_split_indices(
            dataset=train_data,
            n_folds=self.n_folds,
        )
        if self.shuffle_data:
            train_data = train_data.shuffle(seed=self.random_state)

        # iterate over all base estimators
        for i, reg in enumerate(self.base_estimators):
            # iterate over all folds
            for j in range(self.n_folds):
                # build train and test index
                full_idx = list(range(n_samples))
                test_idx = index_lists[j]
                train_idx = list_diff(full_idx, test_idx)
                train_fold = train_data.select(train_idx)
                test_fold = train_data.select(test_idx)

                # train the regressor
                reg.train(train_data=train_fold)

                # generate the new features on the pseudo test set
                reg_tag_name = f"{reg.tag}::{reg.y_features[0]}"
                if self.all_models_support_std:
                    assert supports_std(reg)
                    preds, stds = reg.predict_with_std(test_fold)
                    (
                        new_features[test_idx, i],
                        new_features[test_idx, i + self.n_base_estimators_],
                    ) = (preds[reg_tag_name], stds[f"{reg_tag_name}::std"])
                else:
                    new_features[test_idx, i] = reg.predict(test_fold)[reg_tag_name]

        # build the new dataset for training
        train_comb = copy(train_data)
        meta_estimator_colnames = [clf.tag for clf in self.base_estimators]
        for i_col, colname in enumerate(meta_estimator_colnames):
            train_comb = train_comb.add_column(
                colname,
                new_features[:, i_col],
            ).flatten_indices()
            if self.all_models_support_std:
                train_comb = train_comb.add_column(
                    f"{colname}::std",
                    new_features[:, self.n_base_estimators_ + i_col],
                ).flatten_indices()

        if self.keep_original:
            meta_estimator_colnames += self.x_features
        assert set(meta_estimator_colnames).issubset(
            set(self.meta_estimator.x_features),
        )
        self.meta_estimator.train(train_data=train_comb)

        # train all base classifiers on the full train dataset
        # iterate over all base estimators
        for reg in self.base_estimators:
            reg.train(train_data=train_data)

        return

    def _process_data(self, dataset: Dataset, with_std: bool = False) -> Any:
        """Internal class for `predict`
        Parameters
        ----------
        dataset: Dataset (n_samples, n_features)
            The input samples.
        with_std: bool
            Use predict_with_std or just predict.
        Returns
        -------
        data_new_comb : Dataset
            The processed dataset
        """
        n_samples = dataset.shape[0]

        # initialize matrix for storing newly generated features
        new_features = np.zeros([n_samples, (1 + with_std) * self.n_base_estimators_])

        # build the new features for unknown samples
        # iterate over all base classifiers
        for i, reg in enumerate(self.base_estimators):
            # generate the new features on the test set
            reg_tag_name = f"{reg.tag}::{reg.y_features[0]}"
            if with_std:
                assert supports_std(reg)
                preds, stds = reg.predict_with_std(dataset)
                new_features[:, i], new_features[:, i + self.n_base_estimators_] = (
                    preds[reg_tag_name],
                    stds[f"{reg_tag_name}::std"],
                )
            else:
                new_features[:, i] = reg.predict(dataset)[reg_tag_name]

        # build the new dataset for unknown samples
        new_colnames = [reg.tag for reg in self.base_estimators]
        data_new_comb = copy(dataset)
        for i_col, colname in enumerate(new_colnames):
            data_new_comb = data_new_comb.add_column(
                colname,
                new_features[:, i_col],
            ).flatten_indices()
            if with_std:
                data_new_comb = data_new_comb.add_column(
                    f"{colname}::std",
                    new_features[:, self.n_base_estimators_ + i_col],
                ).flatten_indices()

        return data_new_comb

    def _predict(self, data: Dataset, **kwargs: Any) -> PredictionResult:
        """Make regression predictions for the provided data.
        Parameters
        ----------
        data: Dataset
        Returns
        -------
        labels : numpy array of shape (n_samples,)
            Predictions for each data sample.
        """

        display_names = self._predict_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X_new_comb = self._process_data(x_data)
        original_prediction_results = self.meta_estimator.predict(X_new_comb)
        return dict(
            zip(
                display_names,
                original_prediction_results.values(),
                strict=False,
            ),
        )

    def _predict_with_std(
        self,
        data: Dataset,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """
        Make regression and std predictions for the provided data.
        """
        assert supports_std(
            self.meta_estimator,
        ), "meta_estimator should support uncertainty via prediction with standard deviation"

        (
            prediction_display_names,
            prediction_std_display_names,
        ) = self._predict_with_std_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_std_display_names
            }

        x_data = pick_features(data, self.x_features)
        X_new_comb = self._process_data(x_data, with_std=self.all_models_support_std)
        (
            original_prediction_results,
            original_prediction_stds,
        ) = self.meta_estimator.predict_with_std(X_new_comb)

        return dict(
            zip(
                prediction_display_names,
                original_prediction_results.values(),
                strict=False,
            ),
        ), dict(
            zip(
                prediction_std_display_names,
                original_prediction_stds.values(),
                strict=False,
            ),
        )

    def _predict_with_prediction_interval(
        self,
        data: Dataset,
        confidence: float,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        assert supports_prediction_interval(
            self.meta_estimator,
        ), "meta_estimator should support uncertainty via prediction with prediction interval"

        (
            prediction_display_names,
            prediction_interval_display_names,
        ) = self._predict_with_prediction_interval_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_interval_display_names
            }

        x_data = pick_features(data, self.x_features)
        X_new_comb = self._process_data(x_data, with_std=self.all_models_support_std)
        (
            original_prediction_results,
            original_prediction_intervals,
        ) = self.meta_estimator.predict_with_prediction_interval(
            X_new_comb,
            confidence=confidence,
        )
        return dict(
            zip(
                prediction_display_names,
                original_prediction_results.values(),
                strict=False,
            ),
        ), dict(
            zip(
                prediction_interval_display_names,
                original_prediction_intervals.values(),
                strict=False,
            ),
        )

    def _sample(
        self,
        data: Dataset,
        n_samples: int,
        **kwargs: Any,
    ) -> PredictionResult:
        assert supports_std(
            self.meta_estimator,
        ), "meta_estimator should support uncertainty via predict with std"

        display_names = self._sample_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        prediction_mean_results, prediction_std_results = self._predict_with_std(data)

        prediction_results: PredictionResult = {}
        for display_name, means, stds in zip(
            display_names,
            prediction_mean_results.values(),
            prediction_std_results.values(),
            strict=False,
        ):
            samples = np.random.normal(means, stds, (n_samples, len(means))).T
            prediction_results[display_name] = samples.tolist()

        return prediction_results

    def as_dir(self, directory: str) -> None:
        """
        Serialises a pre-trained ensemble model in a directory.

        As ensembles have a meta estimator and a list of base estimators, each one is
        saved to a subdirectory, in the format:
        `directory/
            meta_estimator/
                ...
            base_estimators/
                0/
                    ...
                1/
                    ...
                ...
        """

        if self.meta_estimator is None:
            raise NotTrainedError

        if self.base_estimators is None:
            raise NotTrainedError

        meta_estimator_path = os.path.join(directory, "meta_estimator")
        base_estimators_base_path = os.path.join(directory, "base_estimators")

        save_to_store(meta_estimator_path, self.meta_estimator)

        for model_index in range(len(self.base_estimators)):
            base_estimator_path = os.path.join(
                base_estimators_base_path,
                str(model_index),
            )
            save_to_store(base_estimator_path, self.base_estimators[model_index])

    def from_dir(self, directory: str) -> None:
        """Deserialises the backend model object stored in a given directory.

        As ensembles have a meta estimator and a list of base estimators, each one is
        saved to a subdirectory, in the format:
        `directory/
            meta_estimator/
                ...
            base_estimators/
                0/
                    ...
                1/
                    ...
                ...
        """

        # The expected model binary
        meta_estimator_path = os.path.join(directory, "meta_estimator")
        base_estimators_base_path = os.path.join(directory, "base_estimators")

        self.meta_estimator = load_from_store(meta_estimator_path)

        base_estimators = []
        for model_index in range(len(self.base_estimators)):
            base_estimator_path = os.path.join(
                base_estimators_base_path,
                str(model_index),
            )
            base_estimators.append(load_from_store(base_estimator_path))

        self.base_estimators = base_estimators



from molflux.datasets.exceptions import ExtrasDependencyImportError


def test_extras_dependency_import_error():
    """That the error can be instantiated"""
    base_exception = Exception()
    datasets_exception = ExtrasDependencyImportError("custom_type", base_exception)

    assert "molflux[custom_type]" in str(datasets_exception)
